@@ -2,7 +2,7 @@
  * Copyright (c) 2004-2005 The Trustees of Indiana University and Indiana
  *                         University Research and Technology
  *                         Corporation.  All rights reserved.
- * Copyright (c) 2004-2015 The University of Tennessee and The University
+ * Copyright (c) 2004-2017 The University of Tennessee and The University
  *                         of Tennessee Research Foundation.  All rights
  *                         reserved.
  * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart,
@@ -12,7 +12,7 @@
  * Copyright (c) 2009      University of Houston. All rights reserved.
  * Copyright (c) 2013      Los Alamos National Security, LLC. All Rights
  *                         reserved.
- * Copyright (c) 2014-2015 Research Organization for Information Science
+ * Copyright (c) 2014-2016 Research Organization for Information Science
  *                         and Technology (RIST). All rights reserved.
  * $COPYRIGHT$
  *
@@ -167,19 +167,16 @@ int ompi_coll_base_allgather_intra_bruck(const void *sbuf, int scount,
        - copy blocks from shift buffer starting at block [rank] in rbuf.
     */
     if (0 != rank) {
-        ptrdiff_t true_extent, true_lb;
         char *free_buf = NULL, *shift_buf = NULL;
+        ptrdiff_t span, gap;
 
-        err = ompi_datatype_get_true_extent(rdtype, &true_lb, &true_extent);
-        if (MPI_SUCCESS != err) { line = __LINE__; goto err_hndl; }
+        span = opal_datatype_span(&rdtype->super, (int64_t)(size - rank) * rcount, &gap);
 
-        free_buf = (char*) calloc(((true_extent +
-                                    ((ptrdiff_t)(size - rank) * (ptrdiff_t)rcount - 1) * rext)),
-                                  sizeof(char));
+        free_buf = (char*)calloc(span, sizeof(char));
         if (NULL == free_buf) {
             line = __LINE__; err = OMPI_ERR_OUT_OF_RESOURCE; goto err_hndl;
         }
-        shift_buf = free_buf - true_lb;
+        shift_buf = free_buf - gap;
 
         /* 1. copy blocks [0 .. (size - rank - 1)] from rbuf to shift buffer */
         err = ompi_datatype_copy_content_same_ddt(rdtype, ((ptrdiff_t)(size - rank) * (ptrdiff_t)rcount),
@@ -206,6 +203,7 @@ int ompi_coll_base_allgather_intra_bruck(const void *sbuf, int scount,
  err_hndl:
     OPAL_OUTPUT((ompi_coll_base_framework.framework_output,  "%s:%4d\tError occurred %d, rank %2d",
                  __FILE__, line, err, rank));
+    (void)line;  // silence compiler warning
     return err;
 }
 
@@ -342,6 +340,7 @@ ompi_coll_base_allgather_intra_recursivedoubling(const void *sbuf, int scount,
  err_hndl:
     OPAL_OUTPUT((ompi_coll_base_framework.framework_output,  "%s:%4d\tError occurred %d, rank %2d",
                  __FILE__, line, err, rank));
+    (void)line;  // silence compiler warning
     return err;
 }
 
@@ -429,6 +428,7 @@ int ompi_coll_base_allgather_intra_ring(const void *sbuf, int scount,
  err_hndl:
     OPAL_OUTPUT((ompi_coll_base_framework.framework_output,  "%s:%4d\tError occurred %d, rank %2d",
                  __FILE__, line, err, rank));
+    (void)line;  // silence compiler warning
     return err;
 }
 
@@ -602,6 +602,7 @@ ompi_coll_base_allgather_intra_neighborexchange(const void *sbuf, int scount,
  err_hndl:
     OPAL_OUTPUT((ompi_coll_base_framework.framework_output,  "%s:%4d\tError occurred %d, rank %2d",
                  __FILE__, line, err, rank));
+    (void)line;  // silence compiler warning
     return err;
 }
 
@@ -622,6 +623,10 @@ int ompi_coll_base_allgather_intra_two_procs(const void *sbuf, int scount,
     OPAL_OUTPUT((ompi_coll_base_framework.framework_output,
                  "ompi_coll_base_allgather_intra_two_procs rank %d", rank));
 
+    if (2 != ompi_comm_size(comm)) {
+        return MPI_ERR_UNSUPPORTED_OPERATION;
+    }
+
     err = ompi_datatype_get_extent (sdtype, &lb, &sext);
     if (MPI_SUCCESS != err) { line = __LINE__; goto err_hndl; }
 
@@ -661,6 +666,7 @@ int ompi_coll_base_allgather_intra_two_procs(const void *sbuf, int scount,
  err_hndl:
     OPAL_OUTPUT((ompi_coll_base_framework.framework_output, "%s:%4d\tError occurred %d, rank %2d",
                  __FILE__, line, err, rank));
+    (void)line;  // silence compiler warning
     return err;
 }
 
@@ -712,20 +718,20 @@ ompi_coll_base_allgather_intra_basic_linear(const void *sbuf, int scount,
 
     /* Gather and broadcast. */
 
-    err = comm->c_coll.coll_gather(sbuf, scount, sdtype,
+    err = comm->c_coll->coll_gather(sbuf, scount, sdtype,
                                    rbuf, rcount, rdtype,
-                                   0, comm, comm->c_coll.coll_gather_module);
+                                   0, comm, comm->c_coll->coll_gather_module);
     if (MPI_SUCCESS == err) {
         size_t length = (ptrdiff_t)rcount * ompi_comm_size(comm);
         if( length < (size_t)INT_MAX ) {
-            err = comm->c_coll.coll_bcast(rbuf, (ptrdiff_t)rcount * ompi_comm_size(comm), rdtype,
-                                          0, comm, comm->c_coll.coll_bcast_module);
+            err = comm->c_coll->coll_bcast(rbuf, (ptrdiff_t)rcount * ompi_comm_size(comm), rdtype,
+                                          0, comm, comm->c_coll->coll_bcast_module);
         } else {
             ompi_datatype_t* temptype;
             ompi_datatype_create_contiguous(ompi_comm_size(comm), rdtype, &temptype);
             ompi_datatype_commit(&temptype);
-            err = comm->c_coll.coll_bcast(rbuf, rcount, temptype,
-                                          0, comm, comm->c_coll.coll_bcast_module);
+            err = comm->c_coll->coll_bcast(rbuf, rcount, temptype,
+                                          0, comm, comm->c_coll->coll_bcast_module);
             ompi_datatype_destroy(&temptype);
         }
     }
