@@ -17,6 +17,7 @@
 #include "mpi.h"
 #include "ompi/mca/mca.h"
 #include "opal/memoryhooks/memory.h"
+#include "opal/mca/memory/base/base.h"
 #include "ompi/mca/coll/coll.h"
 #include "ompi/request/request.h"
 #include "ompi/mca/pml/pml.h"
@@ -49,6 +50,11 @@ typedef struct mca_coll_hcoll_ops_t {
     int (*hcoll_barrier)(void *);
 } mca_coll_hcoll_ops_t;
 
+typedef struct {
+    opal_free_list_item_t super;
+    dte_data_representation_t type;
+} mca_coll_hcoll_dtype_t;
+OBJ_CLASS_DECLARATION(mca_coll_hcoll_dtype_t);
 
 struct mca_coll_hcoll_component_t {
     /** Base coll component */
@@ -89,6 +95,8 @@ struct mca_coll_hcoll_component_t {
     /* FCA global stuff */
     mca_coll_hcoll_ops_t hcoll_ops;
     opal_free_list_t requests;
+    opal_free_list_t dtypes;
+    int derived_types_support_enabled;
 };
 typedef struct mca_coll_hcoll_component_t mca_coll_hcoll_component_t;
 
@@ -172,7 +180,7 @@ int mca_coll_hcoll_bcast(void *buff, int count,
                         struct ompi_communicator_t *comm,
                         mca_coll_base_module_t *module);
 
-int mca_coll_hcoll_allgather(void *sbuf, int scount,
+int mca_coll_hcoll_allgather(const void *sbuf, int scount,
                             struct ompi_datatype_t *sdtype,
                             void *rbuf, int rcount,
                             struct ompi_datatype_t *rdtype,
@@ -195,7 +203,7 @@ int mca_coll_hcoll_gather(const void *sbuf, int scount,
                           struct ompi_communicator_t *comm,
                           mca_coll_base_module_t *module);
 
-int mca_coll_hcoll_allreduce(void *sbuf, void *rbuf, int count,
+int mca_coll_hcoll_allreduce(const void *sbuf, void *rbuf, int count,
                             struct ompi_datatype_t *dtype,
                             struct ompi_op_t *op,
                             struct ompi_communicator_t *comm,
@@ -215,18 +223,18 @@ int mca_coll_hcoll_alltoall(const void *sbuf, int scount,
                            struct ompi_communicator_t *comm,
                            mca_coll_base_module_t *module);
 
-int mca_coll_hcoll_alltoallv(void *sbuf, int *scounts,
-                            int *sdisps,
+int mca_coll_hcoll_alltoallv(const void *sbuf, const int *scounts,
+                            const int *sdisps,
                             struct ompi_datatype_t *sdtype,
-                            void *rbuf, int *rcounts,
-                            int *rdisps,
+                            void *rbuf, const int *rcounts,
+                            const int *rdisps,
                             struct ompi_datatype_t *rdtype,
                             struct ompi_communicator_t *comm,
                             mca_coll_base_module_t *module);
 
-int mca_coll_hcoll_gatherv(void* sbuf, int scount,
+int mca_coll_hcoll_gatherv(const void* sbuf, int scount,
                             struct ompi_datatype_t *sdtype,
-                            void* rbuf, int *rcounts, int *displs,
+                            void* rbuf, const int *rcounts, const int *displs,
                             struct ompi_datatype_t *rdtype,
                             int root,
                             struct ompi_communicator_t *comm,
@@ -242,7 +250,7 @@ int mca_coll_hcoll_ibcast(void *buff, int count,
                             ompi_request_t** request,
                             mca_coll_base_module_t *module);
 
-int mca_coll_hcoll_iallgather(void *sbuf, int scount,
+int mca_coll_hcoll_iallgather(const void *sbuf, int scount,
                             struct ompi_datatype_t *sdtype,
                             void *rbuf, int rcount,
                             struct ompi_datatype_t *rdtype,
@@ -282,7 +290,8 @@ int mca_coll_hcoll_ialltoall(const void *sbuf, int scount,
                             ompi_request_t **req,
                             mca_coll_base_module_t *module);
 
-int mca_coll_hcoll_ialltoallv(void *sbuf, int *scounts,
+#if HCOLL_API >= HCOLL_VERSION(3,7)
+int mca_coll_hcoll_ialltoallv(const void *sbuf, int *scounts,
                             int *sdisps,
                             struct ompi_datatype_t *sdtype,
                             void *rbuf, int *rcounts,
@@ -291,10 +300,11 @@ int mca_coll_hcoll_ialltoallv(void *sbuf, int *scounts,
                             struct ompi_communicator_t *comm,
                             ompi_request_t **req,
                             mca_coll_base_module_t *module);
+#endif
 
 int mca_coll_hcoll_igatherv(const void* sbuf, int scount,
                             struct ompi_datatype_t *sdtype,
-                            void* rbuf, int *rcounts, int *displs,
+                            void* rbuf, const int *rcounts, const int *displs,
                             struct ompi_datatype_t *rdtype,
                             int root,
                             struct ompi_communicator_t *comm,
