@@ -1,16 +1,22 @@
 /**
-  Copyright (c) 2011 Mellanox Technologies. All rights reserved.
-  $COPYRIGHT$
-
-  Additional copyrights may follow
-
- $HEADER$
+ * Copyright (c) 2011 Mellanox Technologies. All rights reserved.
+ * Copyright (c) 2016      IBM Corporation.  All rights reserved.
+ * Copyright (c) 2017      The University of Tennessee and The University
+ *                         of Tennessee Research Foundation.  All rights
+ *                         reserved.
+ * $COPYRIGHT$
+ *
+ * Additional copyrights may follow
+ *
+ * $HEADER$
  */
 
 #include "ompi_config.h"
 #include "coll_hcoll.h"
+#include "coll_hcoll_dtypes.h"
 
 int hcoll_comm_attr_keyval;
+int hcoll_type_attr_keyval;
 
 /*
  * Initial query function that is invoked during MPI_INIT, allowing
@@ -50,6 +56,32 @@ static void mca_coll_hcoll_module_clear(mca_coll_hcoll_module_t *hcoll_module)
     hcoll_module->previous_iallgatherv = NULL;
     hcoll_module->previous_igatherv    = NULL;
     hcoll_module->previous_ireduce     = NULL;
+    hcoll_module->previous_ialltoall   = NULL;
+    hcoll_module->previous_ialltoallv  = NULL;
+
+    hcoll_module->previous_barrier_module = NULL;
+    hcoll_module->previous_bcast_module      = NULL;
+    hcoll_module->previous_allreduce_module  = NULL;
+    hcoll_module->previous_reduce_module     = NULL;
+    hcoll_module->previous_allgather_module  = NULL;
+    hcoll_module->previous_allgatherv_module = NULL;
+    hcoll_module->previous_gather_module     = NULL;
+    hcoll_module->previous_gatherv_module    = NULL;
+    hcoll_module->previous_alltoall_module   = NULL;
+    hcoll_module->previous_alltoallv_module  = NULL;
+    hcoll_module->previous_alltoallw_module  = NULL;
+    hcoll_module->previous_reduce_scatter_module  = NULL;
+    hcoll_module->previous_ibarrier_module    = NULL;
+    hcoll_module->previous_ibcast_module      = NULL;
+    hcoll_module->previous_iallreduce_module  = NULL;
+    hcoll_module->previous_ireduce_module     = NULL;
+    hcoll_module->previous_iallgather_module  = NULL;
+    hcoll_module->previous_iallgatherv_module = NULL;
+    hcoll_module->previous_igatherv_module    = NULL;
+    hcoll_module->previous_ialltoall_module   = NULL;
+    hcoll_module->previous_ialltoallv_module  = NULL;
+
+
 }
 
 static void mca_coll_hcoll_module_construct(mca_coll_hcoll_module_t *hcoll_module)
@@ -63,6 +95,8 @@ void mca_coll_hcoll_mem_release_cb(void *buf, size_t length,
     hcoll_mem_unmap(buf, length, cbdata, from_alloc);
 }
 
+#define OBJ_RELEASE_IF_NOT_NULL( obj ) if( NULL != (obj) ) OBJ_RELEASE( obj );
+
 static void mca_coll_hcoll_module_destruct(mca_coll_hcoll_module_t *hcoll_module)
 {
     int context_destroyed;
@@ -79,25 +113,25 @@ static void mca_coll_hcoll_module_destruct(mca_coll_hcoll_module_t *hcoll_module
        destroy hcoll context*/
 
     if (hcoll_module->hcoll_context != NULL){
-        OBJ_RELEASE(hcoll_module->previous_barrier_module);
-        OBJ_RELEASE(hcoll_module->previous_bcast_module);
-        OBJ_RELEASE(hcoll_module->previous_allreduce_module);
-        OBJ_RELEASE(hcoll_module->previous_allgather_module);
-        OBJ_RELEASE(hcoll_module->previous_allgatherv_module);
-        OBJ_RELEASE(hcoll_module->previous_gatherv_module);
-        OBJ_RELEASE(hcoll_module->previous_alltoall_module);
-        OBJ_RELEASE(hcoll_module->previous_alltoallv_module);
-        OBJ_RELEASE(hcoll_module->previous_reduce_module);
+        OBJ_RELEASE_IF_NOT_NULL(hcoll_module->previous_barrier_module);
+        OBJ_RELEASE_IF_NOT_NULL(hcoll_module->previous_bcast_module);
+        OBJ_RELEASE_IF_NOT_NULL(hcoll_module->previous_allreduce_module);
+        OBJ_RELEASE_IF_NOT_NULL(hcoll_module->previous_allgather_module);
+        OBJ_RELEASE_IF_NOT_NULL(hcoll_module->previous_allgatherv_module);
+        OBJ_RELEASE_IF_NOT_NULL(hcoll_module->previous_gatherv_module);
+        OBJ_RELEASE_IF_NOT_NULL(hcoll_module->previous_alltoall_module);
+        OBJ_RELEASE_IF_NOT_NULL(hcoll_module->previous_alltoallv_module);
+        OBJ_RELEASE_IF_NOT_NULL(hcoll_module->previous_reduce_module);
 
-        OBJ_RELEASE(hcoll_module->previous_ibarrier_module);
-        OBJ_RELEASE(hcoll_module->previous_ibcast_module);
-        OBJ_RELEASE(hcoll_module->previous_iallreduce_module);
-        OBJ_RELEASE(hcoll_module->previous_iallgather_module);
-        OBJ_RELEASE(hcoll_module->previous_iallgatherv_module);
-        OBJ_RELEASE(hcoll_module->previous_igatherv_module);
-        OBJ_RELEASE(hcoll_module->previous_ialltoall_module);
-        OBJ_RELEASE(hcoll_module->previous_ialltoallv_module);
-        OBJ_RELEASE(hcoll_module->previous_ireduce_module);
+        OBJ_RELEASE_IF_NOT_NULL(hcoll_module->previous_ibarrier_module);
+        OBJ_RELEASE_IF_NOT_NULL(hcoll_module->previous_ibcast_module);
+        OBJ_RELEASE_IF_NOT_NULL(hcoll_module->previous_iallreduce_module);
+        OBJ_RELEASE_IF_NOT_NULL(hcoll_module->previous_iallgather_module);
+        OBJ_RELEASE_IF_NOT_NULL(hcoll_module->previous_iallgatherv_module);
+        OBJ_RELEASE_IF_NOT_NULL(hcoll_module->previous_igatherv_module);
+        OBJ_RELEASE_IF_NOT_NULL(hcoll_module->previous_ialltoall_module);
+        OBJ_RELEASE_IF_NOT_NULL(hcoll_module->previous_ialltoallv_module);
+        OBJ_RELEASE_IF_NOT_NULL(hcoll_module->previous_ireduce_module);
 
         /*
         OBJ_RELEASE(hcoll_module->previous_allgatherv_module);
@@ -107,19 +141,20 @@ static void mca_coll_hcoll_module_destruct(mca_coll_hcoll_module_t *hcoll_module
         OBJ_RELEASE(hcoll_module->previous_reduce_scatter_module);
         OBJ_RELEASE(hcoll_module->previous_reduce_module);
         */
-
+#if !defined(HAVE_HCOLL_CONTEXT_FREE)
         context_destroyed = 0;
         hcoll_destroy_context(hcoll_module->hcoll_context,
                               (rte_grp_handle_t)hcoll_module->comm,
                               &context_destroyed);
+#endif
     }
     mca_coll_hcoll_module_clear(hcoll_module);
 }
 
 #define HCOL_SAVE_PREV_COLL_API(__api) do {\
-    hcoll_module->previous_ ## __api            = comm->c_coll.coll_ ## __api;\
-    hcoll_module->previous_ ## __api ## _module = comm->c_coll.coll_ ## __api ## _module;\
-    if (!comm->c_coll.coll_ ## __api || !comm->c_coll.coll_ ## __api ## _module) {\
+    hcoll_module->previous_ ## __api            = comm->c_coll->coll_ ## __api;\
+    hcoll_module->previous_ ## __api ## _module = comm->c_coll->coll_ ## __api ## _module;\
+    if (!comm->c_coll->coll_ ## __api || !comm->c_coll->coll_ ## __api ## _module) {\
         return OMPI_ERROR;\
     }\
     OBJ_RETAIN(hcoll_module->previous_ ## __api ## _module);\
@@ -168,13 +203,17 @@ static int mca_coll_hcoll_save_coll_handlers(mca_coll_hcoll_module_t *hcoll_modu
 /*
 ** Communicator free callback
 */
-int hcoll_comm_attr_del_fn(MPI_Comm comm, int keyval, void *attr_val, void *extra)
+static int hcoll_comm_attr_del_fn(MPI_Comm comm, int keyval, void *attr_val, void *extra)
 {
 
     mca_coll_hcoll_module_t *hcoll_module;
     hcoll_module = (mca_coll_hcoll_module_t*) attr_val;
 
+#ifdef HAVE_HCOLL_CONTEXT_FREE
+    hcoll_context_free(hcoll_module->hcoll_context, (rte_grp_handle_t)comm);
+#else
     hcoll_group_destroy_notify(hcoll_module->hcoll_context);
+#endif
     return OMPI_SUCCESS;
 
 }
@@ -211,6 +250,10 @@ int mca_coll_hcoll_progress(void)
 }
 
 
+OBJ_CLASS_INSTANCE(mca_coll_hcoll_dtype_t,
+                   opal_free_list_item_t,
+                   NULL,NULL);
+
 /*
  * Invoked when there's a new communicator that has been created.
  * Look at the communicator and decide which set of functions and
@@ -288,6 +331,24 @@ mca_coll_hcoll_comm_query(struct ompi_communicator_t *comm, int *priority)
             HCOL_ERROR("Hcol comm keyval create failed");
             return NULL;
         }
+
+        if (mca_coll_hcoll_component.derived_types_support_enabled) {
+            copy_fn.attr_datatype_copy_fn = (MPI_Type_internal_copy_attr_function *) MPI_TYPE_NULL_COPY_FN;
+            del_fn.attr_datatype_delete_fn = hcoll_type_attr_del_fn;
+            err = ompi_attr_create_keyval(TYPE_ATTR, copy_fn, del_fn, &hcoll_type_attr_keyval, NULL ,0, NULL);
+            if (OMPI_SUCCESS != err) {
+                cm->hcoll_enable = 0;
+                hcoll_finalize();
+                opal_progress_unregister(mca_coll_hcoll_progress);
+                HCOL_ERROR("Hcol type keyval create failed");
+                return NULL;
+            }
+        }
+        OBJ_CONSTRUCT(&cm->dtypes, opal_free_list_t);
+        opal_free_list_init(&cm->dtypes, sizeof(mca_coll_hcoll_dtype_t),
+                            8, OBJ_CLASS(mca_coll_hcoll_dtype_t), 0, 0,
+                            32, -1, 32, NULL, 0, NULL, NULL, NULL);
+
     }
 
     hcoll_module = OBJ_NEW(mca_coll_hcoll_module_t);
@@ -346,8 +407,11 @@ mca_coll_hcoll_comm_query(struct ompi_communicator_t *comm, int *priority)
     hcoll_module->super.coll_gather = /*hcoll_collectives.coll_gather ? mca_coll_hcoll_gather :*/ NULL;
     hcoll_module->super.coll_igatherv = hcoll_collectives.coll_igatherv ? mca_coll_hcoll_igatherv : NULL;
     hcoll_module->super.coll_ialltoall = /*hcoll_collectives.coll_ialltoall ? mca_coll_hcoll_ialltoall : */ NULL;
-    hcoll_module->super.coll_ialltoallv = /*hcoll_collectives.coll_ialltoallv ? mca_coll_hcoll_ialltoallv : */ NULL;
-
+#if HCOLL_API >= HCOLL_VERSION(3,7)
+    hcoll_module->super.coll_ialltoallv = hcoll_collectives.coll_ialltoallv ? mca_coll_hcoll_ialltoallv : NULL;
+#else
+    hcoll_module->super.coll_ialltoallv = NULL;
+#endif
     *priority = cm->hcoll_priority;
     module = &hcoll_module->super;
 
