@@ -8,11 +8,15 @@
  * Copyright (c) 2013      The University of Tennessee and The University
  *                         of Tennessee Research Foundation.  All rights
  *                         reserved.
- * Copyright (c) 2014-2015 Research Organization for Information Science
+ * Copyright (c) 2014-2016 Research Organization for Information Science
  *                         and Technology (RIST). All rights reserved.
  * Copyright (c) 2015      Los Alamos National Security, LLC.  All rights
  *                         reserved.
  * Copyright (c) 2015      Mellanox Technologies. All rights reserved.
+ * Copyright (c) 2017      IBM Corporation.  All rights reserved.
+ * $COPYRIGHT$
+ *
+ * Additional copyrights may follow
  *
  * Author(s): Torsten Hoefler <htor@cs.indiana.edu>
  *
@@ -27,17 +31,18 @@
 int ompi_coll_libnbc_igatherv(const void* sendbuf, int sendcount, MPI_Datatype sendtype,
                               void* recvbuf, const int *recvcounts, const int *displs, MPI_Datatype recvtype,
                               int root, struct ompi_communicator_t *comm, ompi_request_t ** request,
-                              struct mca_coll_base_module_2_1_0_t *module) {
+                              struct mca_coll_base_module_2_2_0_t *module) {
   int rank, p, res;
   MPI_Aint rcvext = 0;
   NBC_Schedule *schedule;
-  char *rbuf, inplace;
+  char *rbuf, inplace = 0;
   NBC_Handle *handle;
   ompi_coll_libnbc_module_t *libnbc_module = (ompi_coll_libnbc_module_t*) module;
 
-  NBC_IN_PLACE(sendbuf, recvbuf, inplace);
-
   rank = ompi_comm_rank (comm);
+  if (root == rank) {
+    NBC_IN_PLACE(sendbuf, recvbuf, inplace);
+  }
   p = ompi_comm_size (comm);
 
   if (rank == root) {
@@ -111,7 +116,7 @@ int ompi_coll_libnbc_igatherv(const void* sendbuf, int sendcount, MPI_Datatype s
 int ompi_coll_libnbc_igatherv_inter (const void* sendbuf, int sendcount, MPI_Datatype sendtype,
 				     void* recvbuf, const int *recvcounts, const int *displs, MPI_Datatype recvtype,
 				     int root, struct ompi_communicator_t *comm, ompi_request_t ** request,
-				     struct mca_coll_base_module_2_1_0_t *module) {
+				     struct mca_coll_base_module_2_2_0_t *module) {
   int res, rsize;
   MPI_Aint rcvext;
   NBC_Schedule *schedule;
