@@ -2,16 +2,16 @@
  * Copyright (c) 2004-2005 The Trustees of Indiana University and Indiana
  *                         University Research and Technology
  *                         Corporation.  All rights reserved.
- * Copyright (c) 2004-2005 The University of Tennessee and The University
+ * Copyright (c) 2004-2017 The University of Tennessee and The University
  *                         of Tennessee Research Foundation.  All rights
  *                         reserved.
  * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart,
  *                         University of Stuttgart.  All rights reserved.
  * Copyright (c) 2004-2005 The Regents of the University of California.
  *                         All rights reserved.
- * Copyright (c) 2008-2015 University of Houston. All rights reserved.
+ * Copyright (c) 2008-2016 University of Houston. All rights reserved.
  * Copyright (c) 2015      Los Alamos National Security, LLC. All rights reserved.
- * Copyright (c) 2015      Research Organization for Information Science
+ * Copyright (c) 2015-2017 Research Organization for Information Science
  *                         and Technology (RIST). All rights reserved.
  * $COPYRIGHT$
  *
@@ -26,6 +26,7 @@
 #include "mpi.h"
 #include "ompi/constants.h"
 #include "ompi/mca/fcoll/fcoll.h"
+#include "ompi/mca/fcoll/base/fcoll_base_coll_array.h"
 #include "ompi/mca/io/ompio/io_ompio.h"
 #include "ompi/mca/io/io.h"
 #include "math.h"
@@ -89,13 +90,13 @@ mca_fcoll_static_file_write_all (mca_io_ompio_file_t *fh,
     /* For creating datatype of type io_array */
     int blocklen[3] = {1, 1, 1};
     int static_num_io_procs=1;
-    OPAL_PTRDIFF_TYPE d[3], base;
+    ptrdiff_t d[3], base;
     ompi_datatype_t *types[3];
     ompi_datatype_t *io_array_type=MPI_DATATYPE_NULL;
     int my_aggregator=-1;
     bool sendbuf_is_contiguous= false;
     size_t ftype_size;
-    OPAL_PTRDIFF_TYPE ftype_extent, lb;
+    ptrdiff_t ftype_extent, lb;
 
 
     /*----------------------------------------------*/
@@ -103,7 +104,7 @@ mca_fcoll_static_file_write_all (mca_io_ompio_file_t *fh,
     double write_time = 0.0, start_write_time = 0.0, end_write_time = 0.0;
     double comm_time = 0.0, start_comm_time = 0.0, end_comm_time = 0.0;
     double exch_write = 0.0, start_exch = 0.0, end_exch = 0.0;
-    mca_io_ompio_print_entry nentry;
+    mca_common_ompio_print_entry nentry;
 #endif
 
 
@@ -117,7 +118,7 @@ mca_fcoll_static_file_write_all (mca_io_ompio_file_t *fh,
     /**************************************************************************
      ** 1.  In case the data is not contigous in memory, decode it into an iovec
      **************************************************************************/
-    if ( ( ftype_extent == (OPAL_PTRDIFF_TYPE) ftype_size)             &&
+    if ( ( ftype_extent == (ptrdiff_t) ftype_size)             &&
          opal_datatype_is_contiguous_memory_layout(&datatype->super,1) &&
          0 == lb ) {
         sendbuf_is_contiguous = true;
@@ -154,9 +155,9 @@ mca_fcoll_static_file_write_all (mca_io_ompio_file_t *fh,
     types[1] = &ompi_mpi_long.dt;
     types[2] = &ompi_mpi_int.dt;
 
-    d[0] = (OPAL_PTRDIFF_TYPE)&local_iov_array[0];
-    d[1] = (OPAL_PTRDIFF_TYPE)&local_iov_array[0].length;
-    d[2] = (OPAL_PTRDIFF_TYPE)&local_iov_array[0].process_id;
+    d[0] = (ptrdiff_t)&local_iov_array[0];
+    d[1] = (ptrdiff_t)&local_iov_array[0].length;
+    d[2] = (ptrdiff_t)&local_iov_array[0].process_id;
     base = d[0];
     for (i=0 ; i<3 ; i++) {
         d[i] -= base;
@@ -206,13 +207,13 @@ mca_fcoll_static_file_write_all (mca_io_ompio_file_t *fh,
 #if OMPIO_FCOLL_WANT_TIME_BREAKDOWN
     start_exch = MPI_Wtime();
 #endif
-    ret = fh->f_comm->c_coll.coll_allreduce (&local_cycles,
+    ret = fh->f_comm->c_coll->coll_allreduce (&local_cycles,
                                              &cycles,
                                              1,
                                              MPI_INT,
                                              MPI_MAX,
                                              fh->f_comm,
-                                             fh->f_comm->c_coll.coll_allreduce_module);
+                                             fh->f_comm->c_coll->coll_allreduce_module);
 
     if (OMPI_SUCCESS != ret){
         fprintf(stderr,"local cycles allreduce!\n");
@@ -277,14 +278,14 @@ mca_fcoll_static_file_write_all (mca_io_ompio_file_t *fh,
         }
     }
 
-    iovec_count_per_process = (int *) malloc (fh->f_procs_per_group * sizeof(int));
+    iovec_count_per_process = (int *) calloc (fh->f_procs_per_group, sizeof(int));
     if (NULL == iovec_count_per_process){
         opal_output (1, "OUT OF MEMORY\n");
         ret = OMPI_ERR_OUT_OF_RESOURCE;
         goto exit;
     }
 
-    displs = (int *) malloc (fh->f_procs_per_group * sizeof(int));
+    displs = (int *) calloc (fh->f_procs_per_group, sizeof(int));
     if (NULL == displs){
         opal_output (1, "OUT OF MEMORY\n");
         ret = OMPI_ERR_OUT_OF_RESOURCE;
@@ -294,16 +295,16 @@ mca_fcoll_static_file_write_all (mca_io_ompio_file_t *fh,
 #if OMPIO_FCOLL_WANT_TIME_BREAKDOWN
     start_exch = MPI_Wtime();
 #endif
-    ret = fh->f_allgather_array (&iov_size,
-                                 1,
-                                 MPI_INT,
-                                 iovec_count_per_process,
-                                 1,
-                                 MPI_INT,
-                                 fh->f_aggregator_index,
-                                 fh->f_procs_in_group,
-                                 fh->f_procs_per_group,
-                                 fh->f_comm);
+    ret = fcoll_base_coll_allgather_array (&iov_size,
+                                           1,
+                                           MPI_INT,
+                                           iovec_count_per_process,
+                                           1,
+                                           MPI_INT,
+                                           fh->f_aggregator_index,
+                                           fh->f_procs_in_group,
+                                           fh->f_procs_per_group,
+                                           fh->f_comm);
 
     if( OMPI_SUCCESS != ret){
         fprintf(stderr,"iov size allgatherv array!\n");
@@ -338,17 +339,17 @@ mca_fcoll_static_file_write_all (mca_io_ompio_file_t *fh,
 #if OMPIO_FCOLL_WANT_TIME_BREAKDOWN
     start_exch = MPI_Wtime();
 #endif
-    ret = fh->f_gatherv_array (local_iov_array,
-                               iov_size,
-                               io_array_type,
-                               global_iov_array,
-                               iovec_count_per_process,
-                               displs,
-                               io_array_type,
-                               fh->f_aggregator_index,
-                               fh->f_procs_in_group,
-                               fh->f_procs_per_group,
-                               fh->f_comm);
+    ret = fcoll_base_coll_gatherv_array (local_iov_array,
+                                         iov_size,
+                                         io_array_type,
+                                         global_iov_array,
+                                         iovec_count_per_process,
+                                         displs,
+                                         io_array_type,
+                                         fh->f_aggregator_index,
+                                         fh->f_procs_in_group,
+                                         fh->f_procs_per_group,
+                                         fh->f_comm);
     if (OMPI_SUCCESS != ret){
         fprintf(stderr,"global_iov_array gather error!\n");
         goto exit;
@@ -469,15 +470,15 @@ mca_fcoll_static_file_write_all (mca_io_ompio_file_t *fh,
             }
 
         }
-        if (local_cycles > index) {
-            if ((index == local_cycles-1) && (max_data % bytes_per_cycle)) {
-                bytes_to_write_in_cycle = max_data % bytes_per_cycle;
+        if ( index < local_cycles ) {
+            if ((index == local_cycles-1) && (max_data % (bytes_per_cycle/fh->f_procs_per_group)) ) {
+                bytes_to_write_in_cycle = max_data - total_bytes_written;
             }
-            else if (max_data <= bytes_per_cycle) {
+            else if (max_data <= bytes_per_cycle/fh->f_procs_per_group) {
                 bytes_to_write_in_cycle = max_data;
             }
             else {
-                bytes_to_write_in_cycle = bytes_per_cycle;
+                bytes_to_write_in_cycle = bytes_per_cycle/ fh->f_procs_per_group;
             }
         }
         else {
@@ -499,16 +500,16 @@ mca_fcoll_static_file_write_all (mca_io_ompio_file_t *fh,
         start_exch = MPI_Wtime();
 #endif
         /* gather from each process how many bytes each will be sending */
-        ret = fh->f_gather_array (&bytes_to_write_in_cycle,
-                                  1,
-                                  MPI_INT,
-                                  bytes_per_process,
-                                  1,
-                                  MPI_INT,
-                                  fh->f_aggregator_index,
-                                  fh->f_procs_in_group,
-                                  fh->f_procs_per_group,
-                                  fh->f_comm);
+        ret = fcoll_base_coll_gather_array (&bytes_to_write_in_cycle,
+                                            1,
+                                            MPI_INT,
+                                            bytes_per_process,
+                                            1,
+                                            MPI_INT,
+                                            fh->f_aggregator_index,
+                                            fh->f_procs_in_group,
+                                            fh->f_procs_per_group,
+                                            fh->f_comm);
 
         if (OMPI_SUCCESS != ret){
             fprintf(stderr,"bytes_to_write_in_cycle gather error!\n");
@@ -527,8 +528,6 @@ mca_fcoll_static_file_write_all (mca_io_ompio_file_t *fh,
         */
         if (my_aggregator == fh->f_rank) {
             for (i=0;i<fh->f_procs_per_group; i++){
-/*	    printf("bytes_per_process[%d]: %d\n", i, bytes_per_process[i]);
- */
 
 #if DEBUG_ON
                 printf ("%d : bytes_per_process : %d\n",
@@ -788,7 +787,7 @@ mca_fcoll_static_file_write_all (mca_io_ompio_file_t *fh,
             /* allocate a send buffer and copy the data that needs
                to be sent into it in case the data is non-contigous
                in memory */
-            OPAL_PTRDIFF_TYPE mem_address;
+            ptrdiff_t mem_address;
             size_t remaining = 0;
             size_t temp_position = 0;
 
@@ -801,7 +800,7 @@ mca_fcoll_static_file_write_all (mca_io_ompio_file_t *fh,
             remaining = bytes_to_write_in_cycle;
 
             while (remaining) {
-                mem_address = (OPAL_PTRDIFF_TYPE)
+                mem_address = (ptrdiff_t)
                     (decoded_iov[iov_index].iov_base) + current_position;
 
                 if (remaining >=
@@ -835,7 +834,7 @@ mca_fcoll_static_file_write_all (mca_io_ompio_file_t *fh,
                                  MCA_PML_BASE_SEND_STANDARD,
                                  fh->f_comm,
                                  &send_req));
-
+        
         if ( OMPI_SUCCESS != ret ){
             fprintf(stderr,"isend error!\n");
             goto exit;
@@ -915,7 +914,7 @@ mca_fcoll_static_file_write_all (mca_io_ompio_file_t *fh,
             for (i=0 ; i<fh->f_num_of_io_entries ; i++) {
                 printf(" ADDRESS: %p  OFFSET: %ld   LENGTH: %ld\n",
                        fh->f_io_array[i].memory_address,
-                       (OPAL_PTRDIFF_TYPE)fh->f_io_array[i].offset,
+                       (ptrdiff_t)fh->f_io_array[i].offset,
                        fh->f_io_array[i].length);
             }
 #endif
@@ -953,9 +952,9 @@ mca_fcoll_static_file_write_all (mca_io_ompio_file_t *fh,
     else
         nentry.aggregator = 0;
     nentry.nprocs_for_coll = static_num_io_procs;
-    if (!fh->f_full_print_queue(WRITE_PRINT_QUEUE)){
-	fh->f_register_print_entry(WRITE_PRINT_QUEUE,
-				   nentry);
+    if (!mca_common_ompio_full_print_queue(fh->f_coll_write_time)){
+	mca_common_ompio_register_print_entry(fh->f_coll_write_time,
+                                              nentry);
     }
 #endif
 
@@ -1051,6 +1050,11 @@ exit:
         sorted = NULL;
     }
 
+    if (NULL != displs) {
+        free(displs);
+        displs = NULL;
+    }
+
     return ret;
 }
 
