@@ -10,11 +10,13 @@
  *                         University of Stuttgart.  All rights reserved.
  * Copyright (c) 2004-2005 The Regents of the University of California.
  *                         All rights reserved.
- * Copyright (c) 2007-2015 Los Alamos National Security, LLC.  All rights
+ * Copyright (c) 2007-2016 Los Alamos National Security, LLC.  All rights
  *                         reserved.
  * Copyright (c) 2010-2012 Oracle and/or its affiliates.  All rights reserved.
  * Copyright (c) 2011      Sandia National Laboratories. All rights reserved.
- * Copyright (c) 2014 Cisco Systems, Inc.  All rights reserved.
+ * Copyright (c) 2014      Cisco Systems, Inc.  All rights reserved.
+ * Copyright (c) 2016      Research Organization for Information Science
+ *                         and Technology (RIST). All rights reserved.
  * $COPYRIGHT$
  *
  * Additional copyrights may follow
@@ -28,6 +30,7 @@
 #include "pml_ob1_recvfrag.h"
 #include "ompi/peruse/peruse-internal.h"
 #include "ompi/message/message.h"
+#include "ompi/memchecker.h"
 
 /**
  * Single usage request. As we allow recursive calls to recv
@@ -51,6 +54,7 @@ int mca_pml_ob1_irecv_init(void *addr,
     if (NULL == recvreq)
         return OMPI_ERR_TEMP_OUT_OF_RESOURCE;
 
+    recvreq->req_recv.req_base.req_type = MCA_PML_REQUEST_RECV;
     MCA_PML_OB1_RECV_REQUEST_INIT(recvreq,
                                    addr,
                                    count, datatype, src, tag, comm, true);
@@ -59,6 +63,12 @@ int mca_pml_ob1_irecv_init(void *addr,
                              &((recvreq)->req_recv.req_base),
                              PERUSE_RECV);
 
+    /* Work around a leak in start by marking this request as complete. The
+     * problem occured because we do not have a way to differentiate an
+     * inital request and an incomplete pml request in start. This line
+     * allows us to detect this state. */
+    recvreq->req_recv.req_base.req_pml_complete = true;
+
     *request = (ompi_request_t *) recvreq;
     return OMPI_SUCCESS;
 }
@@ -76,6 +86,7 @@ int mca_pml_ob1_irecv(void *addr,
     if (NULL == recvreq)
         return OMPI_ERR_TEMP_OUT_OF_RESOURCE;
 
+    recvreq->req_recv.req_base.req_type = MCA_PML_REQUEST_RECV;
     MCA_PML_OB1_RECV_REQUEST_INIT(recvreq,
                                    addr,
                                    count, datatype, src, tag, comm, false);
@@ -101,17 +112,18 @@ int mca_pml_ob1_recv(void *addr,
     mca_pml_ob1_recv_request_t *recvreq = NULL;
     int rc;
 
-#if !OMPI_ENABLE_THREAD_MULTIPLE
-    recvreq = mca_pml_ob1_recvreq;
-    mca_pml_ob1_recvreq = NULL;
-    if( OPAL_UNLIKELY(NULL == recvreq) )
-#endif  /* !OMPI_ENABLE_THREAD_MULTIPLE */
-        {
-            MCA_PML_OB1_RECV_REQUEST_ALLOC(recvreq);
-            if (NULL == recvreq)
-                return OMPI_ERR_TEMP_OUT_OF_RESOURCE;
-        }
+    if (OPAL_LIKELY(!ompi_mpi_thread_multiple)) {
+        recvreq = mca_pml_ob1_recvreq;
+        mca_pml_ob1_recvreq = NULL;
+    }
 
+    if( OPAL_UNLIKELY(NULL == recvreq) ) {
+        MCA_PML_OB1_RECV_REQUEST_ALLOC(recvreq);
+        if (NULL == recvreq)
+            return OMPI_ERR_TEMP_OUT_OF_RESOURCE;
+    }
+
+    recvreq->req_recv.req_base.req_type = MCA_PML_REQUEST_RECV;
     MCA_PML_OB1_RECV_REQUEST_INIT(recvreq, addr, count, datatype,
                                   src, tag, comm, false);
 
@@ -128,16 +140,23 @@ int mca_pml_ob1_recv(void *addr,
 
     rc = recvreq->req_recv.req_base.req_ompi.req_status.MPI_ERROR;
 
-#if OMPI_ENABLE_THREAD_MULTIPLE
-    MCA_PML_OB1_RECV_REQUEST_RETURN(recvreq);
-#else
-    if( NULL != mca_pml_ob1_recvreq ) {
+    if (recvreq->req_recv.req_base.req_pml_complete) {
+        /* make buffer defined when the request is compeleted,
+           and before releasing the objects. */
+        MEMCHECKER(
+            memchecker_call(&opal_memchecker_base_mem_defined,
+                            recvreq->req_recv.req_base.req_addr,
+                            recvreq->req_recv.req_base.req_count,
+                            recvreq->req_recv.req_base.req_datatype);
+        );
+    }
+
+    if (OPAL_UNLIKELY(ompi_mpi_thread_multiple || NULL != mca_pml_ob1_recvreq)) {
         MCA_PML_OB1_RECV_REQUEST_RETURN(recvreq);
     } else {
         mca_pml_ob1_recv_request_fini (recvreq);
         mca_pml_ob1_recvreq = recvreq;
     }
-#endif
 
     return rc;
 }
@@ -197,7 +216,7 @@ mca_pml_ob1_imrecv( void *buf,
     recvreq->req_pending = false;
     recvreq->req_ack_sent = false;
 
-    MCA_PML_BASE_RECV_START(&recvreq->req_recv.req_base);
+    MCA_PML_BASE_RECV_START(&recvreq->req_recv);
 
     /* Note - sequence number already assigned */
     recvreq->req_recv.req_base.req_sequence = seq;
@@ -289,7 +308,7 @@ mca_pml_ob1_mrecv( void *buf,
     recvreq->req_rdma_idx = 0;
     recvreq->req_pending = false;
 
-    MCA_PML_BASE_RECV_START(&recvreq->req_recv.req_base);
+    MCA_PML_BASE_RECV_START(&recvreq->req_recv);
 
     /* Note - sequence number already assigned */
     recvreq->req_recv.req_base.req_sequence = seq;
