@@ -19,6 +19,8 @@
  * Copyright (c) 2009      Oak Ridge National Labs.  All rights reserved.
  * Copyright (c) 2009-2012 Los Alamos National Security, LLC.
  *                         All rights reserved.
+ * Copyright (c) 2016      Research Organization for Information Science
+ *                         and Technology (RIST). All rights reserved.
  * $COPYRIGHT$
  *
  * Additional copyrights may follow
@@ -176,7 +178,9 @@
 #define OMPI_MPI_WIN_FLAVOR_DYNAMIC 3
 #define OMPI_MPI_WIN_FLAVOR_SHARED 4
 #define OMPI_MPI_WIN_MODEL 11
+#define OMPI_MPI_WIN_SEPARATE 1
 #define OMPI_MPI_WIN_SIZE 8
+#define OMPI_MPI_WIN_UNIFIED 0
 #define OMPI_MPI_WTIME_IS_GLOBAL 3
 #define OMPI_OMPI_COMM_TYPE_BOARD 8
 #define OMPI_OMPI_COMM_TYPE_CLUSTER 11
@@ -213,10 +217,12 @@
 #define OMPI_MPI_COMPLEX8 19
 #define OMPI_MPI_COUNT 72
 #define OMPI_MPI_CXX_BOOL 54
-#define OMPI_MPI_CXX_CPLEX 55
-#define OMPI_MPI_CXX_DBLCPLEX 56
-#define OMPI_MPI_CXX_LDBLCPLEX 57
-#define OMPI_MPI_C_COMPLEX 68
+#define OMPI_MPI_CXX_COMPLEX 55
+#define OMPI_MPI_CXX_DOUBLE_COMPLEX 56
+#define OMPI_MPI_CXX_FLOAT_COMPLEX 55
+#define OMPI_MPI_CXX_LONG_DOUBLE_COMPLEX 57
+#define OMPI_MPI_C_BOOL 68
+#define OMPI_MPI_C_COMPLEX 69
 #define OMPI_MPI_C_DOUBLE_COMPLEX 70
 #define OMPI_MPI_C_FLOAT_COMPLEX 69
 #define OMPI_MPI_C_LONG_DOUBLE_COMPLEX 71
@@ -253,9 +259,10 @@
 #define OMPI_MPI_LOGICAL4 31
 #define OMPI_MPI_LOGICAL8 32
 #define OMPI_MPI_LONG 41
-#define OMPI_MPI_LONGDBL_INT 50
 #define OMPI_MPI_LONG_DOUBLE 47
+#define OMPI_MPI_LONG_DOUBLE_INT 50
 #define OMPI_MPI_LONG_INT 51
+#define OMPI_MPI_LONG_LONG 43
 #define OMPI_MPI_LONG_LONG_INT 43
 #define OMPI_MPI_LOR 7
 #define OMPI_MPI_LXOR 9
@@ -265,6 +272,7 @@
 #define OMPI_MPI_MESSAGE_NULL 0
 #define OMPI_MPI_MIN 2
 #define OMPI_MPI_MINLOC 12
+#define OMPI_MPI_NO_OP 14
 #define OMPI_MPI_OFFSET 67
 #define OMPI_MPI_OP_NULL 0
 #define OMPI_MPI_PACKED 2
@@ -294,7 +302,6 @@
 #define OMPI_MPI_WIN_NULL 0
 
 #if OMPI_PROVIDE_MPI_FILE_INTERFACE
-#define OMPI_MPI_DISPLACEMENT_CURRENT -54278278
 #define OMPI_MPI_MODE_APPEND 128
 #define OMPI_MPI_MODE_CREATE 1
 #define OMPI_MPI_MODE_DELETE_ON_CLOSE 16
@@ -307,6 +314,7 @@
 #define OMPI_MPI_SEEK_CUR 602
 #define OMPI_MPI_SEEK_END 604
 #define OMPI_MPI_SEEK_SET 600
+#define OMPI_MPI_DISPLACEMENT_CURRENT -54278278
 
 #define OMPI_MPI_FILE_NULL 0
 #endif /* OMPI_PROVIDE_MPI_FILE_INTERFACE */
