@@ -60,6 +60,7 @@ int ompit_var_type_to_datatype (mca_base_var_type_t type, MPI_Datatype *datatype
 
         break;
     case MCA_BASE_VAR_TYPE_STRING:
+    case MCA_BASE_VAR_TYPE_VERSION_STRING:
         *datatype = MPI_CHAR;
         break;
     case MCA_BASE_VAR_TYPE_BOOL:
@@ -85,8 +86,6 @@ int ompit_opal_to_mpit_error (int rc)
     }
 
     switch (rc) {
-    case OPAL_SUCCESS:
-        return MPI_SUCCESS;
     case OPAL_ERR_OUT_OF_RESOURCE:
         return MPI_T_ERR_MEMORY;
     case OPAL_ERR_VALUE_OUT_OF_BOUNDS:
