@@ -1,11 +1,14 @@
 /*
  * Copyright (c) 2012-2013 Los Alamos National Security, LLC.
  *                         All rights reserved.
- * Copyright (c) 2013 Cisco Systems, Inc.  All rights reserved.
+ * Copyright (c) 2013      Cisco Systems, Inc.  All rights reserved.
+ * Copyright (c) 2015-2017 Intel, Inc. All rights reserved.
+ * Copyright (c) 2015      Research Organization for Information Science
+ *                         and Technology (RIST). All rights reserved.
  * $COPYRIGHT$
- * 
+ *
  * Additional copyrights may follow
- * 
+ *
  * $HEADER$
  */
 
@@ -15,9 +18,7 @@
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif  /* HAVE_UNISTD_H */
-#ifdef HAVE_STRING_H
 #include <string.h>
-#endif
 #ifdef HAVE_FCNTL_H
 #include <fcntl.h>
 #endif
@@ -34,6 +35,7 @@
 #include "orte/util/session_dir.h"
 #include "orte/util/show_help.h"
 #include "orte/util/nidmap.h"
+#include "orte/util/threads.h"
 
 #include "orte/mca/errmgr/errmgr.h"
 #include "orte/mca/rml/rml.h"
@@ -298,11 +300,13 @@ static void process_opens(int fd, short args, void *cbdata)
 {
     orte_dfs_request_t *dfs = (orte_dfs_request_t*)cbdata;
     int rc;
-    opal_buffer_t *buffer;
-    char *scheme, *host, *filename;
+    opal_buffer_t *buffer = NULL;
+    char *scheme = NULL, *host = NULL, *filename = NULL;
     int v;
     orte_node_t *node, *nptr;
 
+    ORTE_ACQUIRE_OBJECT(dfs);
+
     /* get the scheme to determine if we can process locally or not */
     if (NULL == (scheme = opal_uri_get_scheme(dfs->uri))) {
         OBJ_RELEASE(dfs);
@@ -324,15 +328,15 @@ static void process_opens(int fd, short args, void *cbdata)
         goto complete;
     }
 
+    free(scheme);
+    scheme = NULL;
+
     /* dissect the uri to extract host and filename/path */
     if (NULL == (filename = opal_filename_from_uri(dfs->uri, &host))) {
         goto complete;
     }
     /* if the host is our own, then treat it as a local file */
-    if (NULL == host ||
-        0 == strcmp(host, orte_process_info.nodename) ||
-        0 == strcmp(host, "localhost") ||
-        opal_ifislocal(host)) {
+    if (NULL == host || orte_ifislocal(host)) {
         opal_output_verbose(1, orte_dfs_base_framework.framework_output,
                             "%s file %s on local host",
                             ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
@@ -363,6 +367,9 @@ static void process_opens(int fd, short args, void *cbdata)
                         "%s file %s on host %s daemon %s",
                         ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
                         filename, host, ORTE_NAME_PRINT(&node->daemon->name));
+
+    free(host);
+    host = NULL;
     /* double-check: if it is our local daemon, then we
      * treat this as local
      */
@@ -400,18 +407,21 @@ static void process_opens(int fd, short args, void *cbdata)
         opal_list_remove_item(&requests, &dfs->super);
         goto complete;
     }
-    
+
     opal_output_verbose(1, orte_dfs_base_framework.framework_output,
                         "%s sending open file request to %s file %s",
                         ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
                         ORTE_NAME_PRINT(&node->daemon->name),
                         filename);
+
+    free(filename);
+    filename = NULL;
     /* send it */
-    if (0 > (rc = orte_rml.send_buffer_nb(&node->daemon->name, buffer,
+    if (0 > (rc = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                          &node->daemon->name, buffer,
                                           ORTE_RML_TAG_DFS_CMD,
                                           orte_rml_send_callback, NULL))) {
         ORTE_ERROR_LOG(rc);
-        OBJ_RELEASE(buffer);
         opal_list_remove_item(&requests, &dfs->super);
         goto complete;
     }
@@ -419,6 +429,18 @@ static void process_opens(int fd, short args, void *cbdata)
     return;
 
  complete:
+    if (NULL != buffer) {
+        OBJ_RELEASE(buffer);
+    }
+    if (NULL != scheme) {
+        free(scheme);
+    }
+    if (NULL != host) {
+        free(host);
+    }
+    if (NULL != filename) {
+        free(filename);
+    }
     OBJ_RELEASE(dfs);
 }
 
@@ -446,7 +468,7 @@ static void dfs_open(char *uri,
     dfs->cbdata = cbdata;
 
     /* post it for processing */
-    ORTE_DFS_POST_REQUEST(dfs, process_opens);
+    ORTE_THREADSHIFT(dfs, orte_event_base, process_opens, ORTE_SYS_PRI);
 }
 
 static void process_close(int fd, short args, void *cbdata)
@@ -457,6 +479,8 @@ static void process_close(int fd, short args, void *cbdata)
     opal_buffer_t *buffer;
     int rc;
 
+    ORTE_ACQUIRE_OBJECT(close_dfs);
+
     opal_output_verbose(1, orte_dfs_base_framework.framework_output,
                         "%s closing fd %d",
                         ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
@@ -500,14 +524,15 @@ static void process_close(int fd, short args, void *cbdata)
         ORTE_ERROR_LOG(rc);
         goto complete;
     }
-    
+
     opal_output_verbose(1, orte_dfs_base_framework.framework_output,
                         "%s sending close file request to %s for fd %d",
                         ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
                         ORTE_NAME_PRINT(&trk->host_daemon),
                         trk->local_fd);
     /* send it */
-    if (0 > (rc = orte_rml.send_buffer_nb(&trk->host_daemon, buffer,
+    if (0 > (rc = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                          &trk->host_daemon, buffer,
                                           ORTE_RML_TAG_DFS_CMD,
                                           orte_rml_send_callback, NULL))) {
         ORTE_ERROR_LOG(rc);
@@ -541,7 +566,7 @@ static void dfs_close(int fd,
     dfs->cbdata = cbdata;
 
     /* post it for processing */
-    ORTE_DFS_POST_REQUEST(dfs, process_close);
+    ORTE_THREADSHIFT(dfs, orte_event_base, process_close, ORTE_SYS_PRI);
 }
 
 static void process_sizes(int fd, short args, void *cbdata)
@@ -553,6 +578,8 @@ static void process_sizes(int fd, short args, void *cbdata)
     int rc;
     struct stat buf;
 
+    ORTE_ACQUIRE_OBJECT(size_dfs);
+
     opal_output_verbose(1, orte_dfs_base_framework.framework_output,
                         "%s processing get_size on fd %d",
                         ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
@@ -612,7 +639,8 @@ static void process_sizes(int fd, short args, void *cbdata)
                         ORTE_NAME_PRINT(&trk->host_daemon),
                         trk->local_fd);
     /* send it */
-    if (0 > (rc = orte_rml.send_buffer_nb(&trk->host_daemon, buffer,
+    if (0 > (rc = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                          &trk->host_daemon, buffer,
                                           ORTE_RML_TAG_DFS_CMD,
                                           orte_rml_send_callback, NULL))) {
         ORTE_ERROR_LOG(rc);
@@ -644,7 +672,7 @@ static void dfs_get_file_size(int fd,
     dfs->cbdata = cbdata;
 
     /* post it for processing */
-    ORTE_DFS_POST_REQUEST(dfs, process_sizes);
+    ORTE_THREADSHIFT(dfs, orte_event_base, process_sizes, ORTE_SYS_PRI);
 }
 
 
@@ -658,6 +686,8 @@ static void process_seeks(int fd, short args, void *cbdata)
     int rc;
     struct stat buf;
 
+    ORTE_ACQUIRE_OBJECT(seek_dfs);
+
     opal_output_verbose(1, orte_dfs_base_framework.framework_output,
                         "%s processing seek on fd %d",
                         ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
@@ -760,7 +790,8 @@ static void process_seeks(int fd, short args, void *cbdata)
                         ORTE_NAME_PRINT(&trk->host_daemon),
                         trk->local_fd);
     /* send it */
-    if (0 > (rc = orte_rml.send_buffer_nb(&trk->host_daemon, buffer,
+    if (0 > (rc = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                          &trk->host_daemon, buffer,
                                           ORTE_RML_TAG_DFS_CMD,
                                           orte_rml_send_callback, NULL))) {
         ORTE_ERROR_LOG(rc);
@@ -792,7 +823,7 @@ static void dfs_seek(int fd, long offset, int whence,
     dfs->cbdata = cbdata;
 
     /* post it for processing */
-    ORTE_DFS_POST_REQUEST(dfs, process_seeks);
+    ORTE_THREADSHIFT(dfs, orte_event_base, process_seeks, ORTE_SYS_PRI);
 }
 
 static void process_reads(int fd, short args, void *cbdata)
@@ -805,6 +836,8 @@ static void process_reads(int fd, short args, void *cbdata)
     int64_t i64;
     int rc;
 
+    ORTE_ACQUIRE_OBJECT(read_dfs);
+
     /* look in our local records for this fd */
     trk = NULL;
     for (item = opal_list_get_first(&active_files);
@@ -863,14 +896,15 @@ static void process_reads(int fd, short args, void *cbdata)
         ORTE_ERROR_LOG(rc);
         goto complete;
     }
-    
+
     opal_output_verbose(1, orte_dfs_base_framework.framework_output,
                         "%s sending read file request to %s for fd %d",
                         ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
                         ORTE_NAME_PRINT(&trk->host_daemon),
                         trk->local_fd);
     /* send it */
-    if (0 > (rc = orte_rml.send_buffer_nb(&trk->host_daemon, buffer,
+    if (0 > (rc = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                          &trk->host_daemon, buffer,
                                           ORTE_RML_TAG_DFS_CMD,
                                           orte_rml_send_callback, NULL))) {
         ORTE_ERROR_LOG(rc);
@@ -901,7 +935,7 @@ static void dfs_read(int fd, uint8_t *buffer,
     dfs->cbdata = cbdata;
 
     /* post it for processing */
-    ORTE_DFS_POST_REQUEST(dfs, process_reads);
+    ORTE_THREADSHIFT(dfs, orte_event_base, process_reads, ORTE_SYS_PRI);
 }
 
 static void process_posts(int fd, short args, void *cbdata)
@@ -912,6 +946,8 @@ static void process_posts(int fd, short args, void *cbdata)
     opal_list_item_t *item;
     int rc;
 
+    ORTE_ACQUIRE_OBJECT(dfs);
+
     opal_output_verbose(1, orte_dfs_base_framework.framework_output,
                         "%s posting file map containing %d bytes for target %s",
                         ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
@@ -986,7 +1022,7 @@ static void dfs_post_file_map(opal_buffer_t *buffer,
     dfs->cbdata = cbdata;
 
     /* post it for processing */
-    ORTE_DFS_POST_REQUEST(dfs, process_posts);
+    ORTE_THREADSHIFT(dfs, orte_event_base, process_posts, ORTE_SYS_PRI);
 }
 
 static int get_job_maps(orte_dfs_jobfm_t *jfm,
@@ -1034,6 +1070,8 @@ static void process_getfm(int fd, short args, void *cbdata)
     int32_t n, ntotal;
     int rc;
 
+    ORTE_ACQUIRE_OBJECT(dfs);
+
     /* if the target job is WILDCARD, then process
      * data for all jobids - else, find the one
      */
@@ -1097,7 +1135,7 @@ static void dfs_get_file_map(orte_process_name_t *target,
     dfs->cbdata = cbdata;
 
     /* post it for processing */
-    ORTE_DFS_POST_REQUEST(dfs, process_getfm);
+    ORTE_THREADSHIFT(dfs, orte_event_base, process_getfm, ORTE_SYS_PRI);
 }
 
 static void process_load(int fd, short args, void *cbdata)
@@ -1112,6 +1150,8 @@ static void process_load(int fd, short args, void *cbdata)
     int rc;
     opal_buffer_t *xfer;
 
+    ORTE_ACQUIRE_OBJECT(dfs);
+
     /* see if we already have a tracker for this job */
     jfm = NULL;
     for (item = opal_list_get_first(&file_maps);
@@ -1210,7 +1250,7 @@ static void dfs_load_file_maps(orte_jobid_t jobid,
     dfs->cbdata = cbdata;
 
     /* post it for processing */
-    ORTE_DFS_POST_REQUEST(dfs, process_load);
+    ORTE_THREADSHIFT(dfs, orte_event_base, process_load, ORTE_SYS_PRI);
 }
 
 static void process_purge(int fd, short args, void *cbdata)
@@ -1219,6 +1259,8 @@ static void process_purge(int fd, short args, void *cbdata)
     opal_list_item_t *item;
     orte_dfs_jobfm_t *jfm, *jptr;
 
+    ORTE_ACQUIRE_OBJECT(dfs);
+
     /* find the job tracker */
     jfm = NULL;
     for (item = opal_list_get_first(&file_maps);
@@ -1265,7 +1307,7 @@ static void dfs_purge_file_maps(orte_jobid_t jobid,
     dfs->cbdata = cbdata;
 
     /* post it for processing */
-    ORTE_DFS_POST_REQUEST(dfs, process_purge);
+    ORTE_THREADSHIFT(dfs, orte_event_base, process_purge, ORTE_SYS_PRI);
 }
 
 
@@ -1367,7 +1409,8 @@ static void recv_dfs_cmd(int status, orte_process_name_t* sender,
             return;
         }
         /* send it */
-        if (0 > (rc = orte_rml.send_buffer_nb(sender, answer,
+        if (0 > (rc = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                              sender, answer,
                                               ORTE_RML_TAG_DFS_DATA,
                                               orte_rml_send_callback, NULL))) {
             ORTE_ERROR_LOG(rc);
@@ -1455,7 +1498,8 @@ static void recv_dfs_cmd(int status, orte_process_name_t* sender,
             return;
         }
         /* send it */
-        if (0 > (rc = orte_rml.send_buffer_nb(sender, answer,
+        if (0 > (rc = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                              sender, answer,
                                               ORTE_RML_TAG_DFS_DATA,
                                               orte_rml_send_callback, NULL))) {
             ORTE_ERROR_LOG(rc);
@@ -1564,7 +1608,8 @@ static void recv_dfs_cmd(int status, orte_process_name_t* sender,
                             ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
                             (long)bytes_read,
                             ORTE_NAME_PRINT(sender));
-        if (0 > (rc = orte_rml.send_buffer_nb(sender, answer,
+        if (0 > (rc = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                              sender, answer,
                                               ORTE_RML_TAG_DFS_DATA,
                                               orte_rml_send_callback, NULL))) {
             ORTE_ERROR_LOG(rc);
@@ -1619,6 +1664,10 @@ static void recv_dfs_cmd(int status, orte_process_name_t* sender,
                                         (long)i64, my_fd);
                     /* do the read */
                     read_buf = (uint8_t*)malloc(i64);
+                    if (NULL == read_buf) {
+                        ORTE_ERROR_LOG(ORTE_ERR_OUT_OF_RESOURCE);
+                        goto answer_read;
+                    }
                     bytes_read = read(my_fd, read_buf, (long)i64);
                     if (0 < bytes_read) {
                         /* update our location */
@@ -1633,40 +1682,55 @@ static void recv_dfs_cmd(int status, orte_process_name_t* sender,
         answer = OBJ_NEW(opal_buffer_t);
         if (OPAL_SUCCESS != (rc = opal_dss.pack(answer, &cmd, 1, ORTE_DFS_CMD_T))) {
             ORTE_ERROR_LOG(rc);
+            OBJ_RELEASE(answer);
+            if (NULL != read_buf) {
+                free(read_buf);
+            }
             return;
         }
         if (OPAL_SUCCESS != (rc = opal_dss.pack(answer, &rid, 1, OPAL_UINT64))) {
             ORTE_ERROR_LOG(rc);
+            OBJ_RELEASE(answer);
+            if (NULL != read_buf) {
+                free(read_buf);
+            }
             return;
         }
         /* include the number of bytes read */
         if (OPAL_SUCCESS != (rc = opal_dss.pack(answer, &bytes_read, 1, OPAL_INT64))) {
             ORTE_ERROR_LOG(rc);
+            OBJ_RELEASE(answer);
+            if (NULL != read_buf) {
+                free(read_buf);
+            }
             return;
         }
         /* include the bytes read */
         if (0 < bytes_read) {
             if (OPAL_SUCCESS != (rc = opal_dss.pack(answer, read_buf, bytes_read, OPAL_UINT8))) {
                 ORTE_ERROR_LOG(rc);
+                OBJ_RELEASE(answer);
+                free(read_buf);
                 return;
             }
         }
+        if (NULL != read_buf) {
+            free(read_buf);
+        }
         /* send it */
         opal_output_verbose(1, orte_dfs_base_framework.framework_output,
                             "%s sending %ld bytes back to %s",
                             ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
                             (long)bytes_read,
                             ORTE_NAME_PRINT(sender));
-        if (0 > (rc = orte_rml.send_buffer_nb(sender, answer,
+        if (0 > (rc = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                              sender, answer,
                                               ORTE_RML_TAG_DFS_DATA,
                                               orte_rml_send_callback, NULL))) {
             ORTE_ERROR_LOG(rc);
             OBJ_RELEASE(answer);
             return;
         }
-        if (NULL != read_buf) {
-            free(read_buf);
-        }
         break;
 
     case ORTE_DFS_POST_CMD:
@@ -1712,7 +1776,8 @@ static void recv_dfs_cmd(int status, orte_process_name_t* sender,
                 ORTE_ERROR_LOG(rc);
                 return;
             }
-            if (0 > (rc = orte_rml.send_buffer_nb(sender, answer,
+            if (0 > (rc = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                                  sender, answer,
                                                   ORTE_RML_TAG_DFS_DATA,
                                                   orte_rml_send_callback, NULL))) {
                 ORTE_ERROR_LOG(rc);
@@ -1881,7 +1946,8 @@ static void recv_dfs_cmd(int status, orte_process_name_t* sender,
                             "%s getf-cmd: returning %d maps with %d bytes to sender %s",
                             ORTE_NAME_PRINT(ORTE_PROC_MY_NAME), nmaps,
                             (int)answer->bytes_used, ORTE_NAME_PRINT(sender));
-        if (0 > (rc = orte_rml.send_buffer_nb(sender, answer,
+        if (0 > (rc = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                              sender, answer,
                                               ORTE_RML_TAG_DFS_DATA,
                                               orte_rml_send_callback, NULL))) {
             ORTE_ERROR_LOG(rc);
@@ -2136,7 +2202,8 @@ static void remote_open(int fd, short args, void *cbdata)
         return;
     }
     /* send it */
-    if (0 > (rc = orte_rml.send_buffer_nb(&req->trk->requestor, answer,
+    if (0 > (rc = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                          &req->trk->requestor, answer,
                                           ORTE_RML_TAG_DFS_DATA,
                                           orte_rml_send_callback, NULL))) {
         ORTE_ERROR_LOG(rc);
@@ -2177,7 +2244,8 @@ static void remote_size(int fd, short args, void *cbdata)
         return;
     }
     /* send it */
-    if (0 > (rc = orte_rml.send_buffer_nb(&req->trk->requestor, answer,
+    if (0 > (rc = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                          &req->trk->requestor, answer,
                                           ORTE_RML_TAG_DFS_DATA,
                                           orte_rml_send_callback, NULL))) {
         ORTE_ERROR_LOG(rc);
@@ -2241,7 +2309,8 @@ static void remote_seek(int fd, short args, void *cbdata)
         return;
     }
     /* send it */
-    if (0 > (rc = orte_rml.send_buffer_nb(&req->trk->requestor, answer,
+    if (0 > (rc = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                          &req->trk->requestor, answer,
                                           ORTE_RML_TAG_DFS_DATA,
                                           orte_rml_send_callback, NULL))) {
         ORTE_ERROR_LOG(rc);
@@ -2263,6 +2332,10 @@ static void remote_read(int fd, short args, void *cbdata)
                         "%s issuing read",
                         ORTE_NAME_PRINT(ORTE_PROC_MY_NAME));
     read_buf = (uint8_t*)malloc(req->nbytes);
+    if (NULL == read_buf) {
+        ORTE_ERROR_LOG(ORTE_ERR_OUT_OF_RESOURCE);
+        return;
+    }
     bytes_read = read(req->trk->local_fd, read_buf, (long)req->nbytes);
     if (0 < bytes_read) {
         /* update our location */
@@ -2272,38 +2345,45 @@ static void remote_read(int fd, short args, void *cbdata)
     answer = OBJ_NEW(opal_buffer_t);
     if (OPAL_SUCCESS != (rc = opal_dss.pack(answer, &cmd, 1, ORTE_DFS_CMD_T))) {
         ORTE_ERROR_LOG(rc);
+        free(read_buf);
         return;
     }
     if (OPAL_SUCCESS != (rc = opal_dss.pack(answer, &req->rid, 1, OPAL_UINT64))) {
         ORTE_ERROR_LOG(rc);
+        free(read_buf);
+        OBJ_RELEASE(answer);
         return;
     }
     /* include the number of bytes read */
     if (OPAL_SUCCESS != (rc = opal_dss.pack(answer, &bytes_read, 1, OPAL_INT64))) {
         ORTE_ERROR_LOG(rc);
+        free(read_buf);
+        OBJ_RELEASE(answer);
         return;
     }
     /* include the bytes read */
     if (0 < bytes_read) {
         if (OPAL_SUCCESS != (rc = opal_dss.pack(answer, read_buf, bytes_read, OPAL_UINT8))) {
             ORTE_ERROR_LOG(rc);
+            free(read_buf);
+            OBJ_RELEASE(answer);
             return;
         }
     }
+    free(read_buf);
     /* send it */
     opal_output_verbose(1, orte_dfs_base_framework.framework_output,
                         "%s sending %ld bytes back to %s",
                         ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
                         (long)bytes_read,
                         ORTE_NAME_PRINT(&req->trk->requestor));
-    if (0 > (rc = orte_rml.send_buffer_nb(&req->trk->requestor, answer,
+    if (0 > (rc = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                          &req->trk->requestor, answer,
                                           ORTE_RML_TAG_DFS_DATA,
                                           orte_rml_send_callback, NULL))) {
         ORTE_ERROR_LOG(rc);
         OBJ_RELEASE(answer);
         return;
     }
-    free(read_buf);
     OBJ_RELEASE(req);
 }
-
