@@ -5,14 +5,16 @@
  * Copyright (c) 2004-2005 The University of Tennessee and The University
  *                         of Tennessee Research Foundation.  All rights
  *                         reserved.
- * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart, 
+ * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart,
  *                         University of Stuttgart.  All rights reserved.
  * Copyright (c) 2004-2005 The Regents of the University of California.
  *                         All rights reserved.
+ * Copyright (c) 2008      Cisco Systems, Inc.  All rights reserved.
+ * Copyright (c) 2016-2017 Intel, Inc.  All rights reserved.
  * $COPYRIGHT$
- * 
+ *
  * Additional copyrights may follow
- * 
+ *
  * $HEADER$
  *
  */
@@ -20,7 +22,10 @@
 #ifndef IOF_BASE_SETUP_H_
 #define IOF_BASE_SETUP_H_
 
-#include "orte/mca/ns/ns.h"
+#include "orte_config.h"
+#include "orte/types.h"
+
+#include "orte/mca/iof/base/base.h"
 
 struct orte_iof_base_io_conf_t {
     int usepty;
@@ -30,6 +35,7 @@ struct orte_iof_base_io_conf_t {
     int p_stdin[2];
     int p_stdout[2];
     int p_stderr[2];
+    int p_internal[2];
 };
 typedef struct orte_iof_base_io_conf_t orte_iof_base_io_conf_t;
 
@@ -42,9 +48,15 @@ typedef struct orte_iof_base_io_conf_t orte_iof_base_io_conf_t;
  */
 ORTE_DECLSPEC int orte_iof_base_setup_prefork(orte_iof_base_io_conf_t *opts);
 
-ORTE_DECLSPEC int orte_iof_base_setup_child(orte_iof_base_io_conf_t *opts);
+ORTE_DECLSPEC int orte_iof_base_setup_child(orte_iof_base_io_conf_t *opts,
+                                            char ***env);
 
 ORTE_DECLSPEC int orte_iof_base_setup_parent(const orte_process_name_t* name,
                                              orte_iof_base_io_conf_t *opts);
 
+/* setup output files */
+ORTE_DECLSPEC int orte_iof_base_setup_output_files(const orte_process_name_t* dst_name,
+                                                   orte_job_t *jobdat,
+                                                   orte_iof_proc_t *proct);
+
 #endif
