@@ -2,6 +2,9 @@
  * Copyright (c) 2011      Cisco Systems, Inc.  All rights reserved.
  * Copyright (c) 2011      Los Alamos National Security, LLC.
  *                         All rights reserved.
+ * Copyright (c) 2014-2017 Intel, Inc.  All rights reserved.
+ * Copyright (c) 2015      Research Organization for Information Science
+ *                         and Technology (RIST). All rights reserved.
  * $COPYRIGHT$
  *
  * Additional copyrights may follow
@@ -17,9 +20,7 @@
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif  /* HAVE_UNISTD_H */
-#ifdef HAVE_STRING_H
 #include <string.h>
-#endif  /* HAVE_STRING_H */
 
 #include "opal/mca/hwloc/base/base.h"
 #include "opal/util/argv.h"
@@ -32,9 +33,11 @@
 #include "rmaps_ppr.h"
 
 static int ppr_mapper(orte_job_t *jdata);
+static int assign_locations(orte_job_t *jdata);
 
 orte_rmaps_base_module_t orte_rmaps_ppr_module = {
-    ppr_mapper
+    .map_job = ppr_mapper,
+    .assign_locations = assign_locations
 };
 
 /* RHC: will eventually remove this
@@ -57,13 +60,11 @@ typedef enum {
     OPAL_HWLOC_HWTHREAD_LEVEL
 } opal_hwloc_level_t;
 
-#if OPAL_HAVE_HWLOC
 static void prune(orte_jobid_t jobid,
                   orte_app_idx_t app_idx,
                   orte_node_t *node,
                   opal_hwloc_level_t *level,
                   orte_vpid_t *nmapped);
-#endif
 
 static int ppr[OPAL_HWLOC_HWTHREAD_LEVEL+1];
 
@@ -76,14 +77,12 @@ static int ppr_mapper(orte_job_t *jdata)
     orte_app_context_t *app;
     orte_vpid_t total_procs, nprocs_mapped;
     opal_hwloc_level_t start=OPAL_HWLOC_NODE_LEVEL;
-#if OPAL_HAVE_HWLOC
     hwloc_obj_t obj;
     hwloc_obj_type_t lowest;
     unsigned cache_level=0;
     unsigned int nobjs, i;
     bool pruning_reqd = false;
     opal_hwloc_level_t level;
-#endif
     opal_list_t node_list;
     opal_list_item_t *item;
     orte_std_cntr_t num_slots;
@@ -95,7 +94,7 @@ static int ppr_mapper(orte_job_t *jdata)
     /* only handle initial launch of loadbalanced
      * or NPERxxx jobs - allow restarting of failed apps
      */
-    if (ORTE_JOB_CONTROL_RESTART & jdata->controls) {
+    if (ORTE_FLAG_TEST(jdata, ORTE_JOB_FLAG_RESTART)) {
         opal_output_verbose(5, orte_rmaps_base_framework.framework_output,
                             "mca:rmaps:ppr: job %s being restarted - ppr cannot map",
                             ORTE_JOBID_PRINT(jdata->jobid));
@@ -113,8 +112,10 @@ static int ppr_mapper(orte_job_t *jdata)
         ORTE_MAPPING_PPR != ORTE_GET_MAPPING_POLICY(jdata->map->mapping)) {
         /* not for us */
         opal_output_verbose(5, orte_rmaps_base_framework.framework_output,
-                            "mca:rmaps:ppr: job %s not using ppr mapper",
-                            ORTE_JOBID_PRINT(jdata->jobid));
+                            "mca:rmaps:ppr: job %s not using ppr mapper PPR %s policy %s",
+                            ORTE_JOBID_PRINT(jdata->jobid),
+                            (NULL == jdata->map->ppr) ? "NULL" : jdata->map->ppr,
+                            (ORTE_MAPPING_PPR == ORTE_GET_MAPPING_POLICY(jdata->map->mapping)) ? "PPRSET" : "PPR NOTSET");
         return ORTE_ERR_TAKE_NEXT_OPTION;
     }
 
@@ -150,7 +151,6 @@ static int ppr_mapper(orte_job_t *jdata)
             ORTE_SET_MAPPING_POLICY(jdata->map->mapping, ORTE_MAPPING_BYNODE);
             start = OPAL_HWLOC_NODE_LEVEL;
             n++;
-#if OPAL_HAVE_HWLOC
         } else if (0 == strncasecmp(ck[1], "hwthread", len) ||
                    0 == strncasecmp(ck[1], "thread", len)) {
             ppr[OPAL_HWLOC_HWTHREAD_LEVEL] = strtol(ck[0], NULL, 10);
@@ -203,7 +203,6 @@ static int ppr_mapper(orte_job_t *jdata)
                 ORTE_SET_MAPPING_POLICY(jdata->map->mapping, ORTE_MAPPING_BYNUMA);
             }
             n++;
-#endif
         } else {
             /* unknown spec */
             orte_show_help("help-orte-rmaps-ppr.txt", "unrecognized-ppr-option", true, ck[1], jdata->map->ppr);
@@ -219,23 +218,19 @@ static int ppr_mapper(orte_job_t *jdata)
         opal_output(0, "NOTHING GIVEN");
         return ORTE_ERR_SILENT;
     }
-#if OPAL_HAVE_HWLOC
     /* if more than one level was specified, then pruning will be reqd */
     if (1 < n) {
         pruning_reqd = true;
     }
-#endif
 
     opal_output_verbose(5, orte_rmaps_base_framework.framework_output,
                         "mca:rmaps:ppr: job %s assigned policy %s",
                         ORTE_JOBID_PRINT(jdata->jobid),
                         orte_rmaps_base_print_mapping(jdata->map->mapping));
 
-#if OPAL_HAVE_HWLOC
     /* convenience */
     level = start;
     lowest = opal_hwloc_levels[start];
-#endif
 
     for (idx=0; idx < (orte_app_idx_t)jdata->apps->size; idx++) {
         if (NULL == (app = (orte_app_context_t*)opal_pointer_array_get_item(jdata->apps, idx))) {
@@ -271,46 +266,34 @@ static int ppr_mapper(orte_job_t *jdata)
              item != opal_list_get_end(&node_list);
              item = opal_list_get_next(item)) {
             node = (orte_node_t*)item;
-#if OPAL_HAVE_HWLOC
             /* bozo check */
-            if (NULL == node->topology) {
+            if (NULL == node->topology || NULL == node->topology->topo) {
                 orte_show_help("help-orte-rmaps-ppr.txt", "ppr-topo-missing",
                                true, node->name);
                 rc = ORTE_ERR_SILENT;
                 goto error;
             }
-#endif
             /* add the node to the map, if needed */
-            if (!node->mapped) {
-                if (ORTE_SUCCESS > (rc = opal_pointer_array_add(jdata->map->nodes, (void*)node))) {
-                    ORTE_ERROR_LOG(rc);
-                    goto error;
-                }
-                node->mapped = true;
-                OBJ_RETAIN(node);  /* maintain accounting on object */
+            if (!ORTE_FLAG_TEST(node, ORTE_NODE_FLAG_MAPPED)) {
+                ORTE_FLAG_SET(node, ORTE_NODE_FLAG_MAPPED);
                 jdata->map->num_nodes++;
             }
             /* if we are mapping solely at the node level, just put
              * that many procs on this node
              */
             if (OPAL_HWLOC_NODE_LEVEL == start) {
-#if OPAL_HAVE_HWLOC
-                obj = hwloc_get_root_obj(node->topology);
-#endif
+                obj = hwloc_get_root_obj(node->topology->topo);
                 for (j=0; j < ppr[start] && nprocs_mapped < total_procs; j++) {
                     if (NULL == (proc = orte_rmaps_base_setup_proc(jdata, node, idx))) {
                         rc = ORTE_ERR_OUT_OF_RESOURCE;
                         goto error;
                     }
                     nprocs_mapped++;
-#if OPAL_HAVE_HWLOC
-                    proc->locale = obj;
-#endif
+                    orte_set_attribute(&proc->attributes, ORTE_PROC_HWLOC_LOCALE, ORTE_ATTR_LOCAL, obj, OPAL_PTR);
                 }
-#if OPAL_HAVE_HWLOC
             } else {
                 /* get the number of lowest resources on this node */
-                nobjs = opal_hwloc_base_get_nbobjs_by_type(node->topology,
+                nobjs = opal_hwloc_base_get_nbobjs_by_type(node->topology->topo,
                                                            lowest, cache_level,
                                                            OPAL_HWLOC_AVAILABLE);
 
@@ -318,7 +301,7 @@ static int ppr_mapper(orte_job_t *jdata)
                  * recording the locale of each proc so we know its cpuset
                  */
                 for (i=0; i < nobjs; i++) {
-                    obj = opal_hwloc_base_get_obj_by_type(node->topology,
+                    obj = opal_hwloc_base_get_obj_by_type(node->topology->topo,
                                                           lowest, cache_level,
                                                           i, OPAL_HWLOC_AVAILABLE);
                     for (j=0; j < ppr[start] && nprocs_mapped < total_procs; j++) {
@@ -327,7 +310,7 @@ static int ppr_mapper(orte_job_t *jdata)
                             goto error;
                         }
                         nprocs_mapped++;
-                        proc->locale = obj;
+                        orte_set_attribute(&proc->attributes, ORTE_PROC_HWLOC_LOCALE, ORTE_ATTR_LOCAL, obj, OPAL_PTR);
                     }
                 }
 
@@ -339,46 +322,51 @@ static int ppr_mapper(orte_job_t *jdata)
                     level--;
                     prune(jdata->jobid, idx, node, &level, &nprocs_mapped);
                 }
-#endif
             }
 
-            /* set the total slots used */
-            if ((int)node->num_procs <= node->slots) {
-                node->slots_inuse = (int)node->num_procs;
-            } else {
-                node->slots_inuse = node->slots;
-            }
-
-            /* if no-oversubscribe was specified, check to see if
-             * we have violated the total slot specification - regardless,
-             * if slots_max was given, we are not allowed to violate it!
-             */
-            if ((node->slots < (int)node->num_procs) ||
-                (0 < node->slots_max && node->slots_max < (int)node->num_procs)) {
-                if (ORTE_MAPPING_NO_OVERSUBSCRIBE & ORTE_GET_MAPPING_DIRECTIVE(jdata->map->mapping)) {
-                    orte_show_help("help-orte-rmaps-base.txt", "orte-rmaps-base:alloc-error",
-                                   true, node->num_procs, app->app);
-                    rc = ORTE_ERR_SILENT;
-                    goto error;
+            if (!(ORTE_MAPPING_DEBUGGER & ORTE_GET_MAPPING_DIRECTIVE(jdata->map->mapping))) {
+                /* set the total slots used */
+                if ((int)node->num_procs <= node->slots) {
+                    node->slots_inuse = (int)node->num_procs;
+                } else {
+                    node->slots_inuse = node->slots;
                 }
-                /* flag the node as oversubscribed so that sched-yield gets
-                 * properly set
+
+                /* if no-oversubscribe was specified, check to see if
+                 * we have violated the total slot specification - regardless,
+                 * if slots_max was given, we are not allowed to violate it!
                  */
-                node->oversubscribed = true;
-                /* check for permission */
-                if (node->slots_given) {
-                    /* if we weren't given a directive either way, then we will error out
-                     * as the #slots were specifically given, either by the host RM or
-                     * via hostfile/dash-host */
-                    if (!(ORTE_MAPPING_SUBSCRIBE_GIVEN & ORTE_GET_MAPPING_DIRECTIVE(orte_rmaps_base.mapping))) {
+                if ((node->slots < (int)node->num_procs) ||
+                    (0 < node->slots_max && node->slots_max < (int)node->num_procs)) {
+                    if (ORTE_MAPPING_NO_OVERSUBSCRIBE & ORTE_GET_MAPPING_DIRECTIVE(jdata->map->mapping)) {
                         orte_show_help("help-orte-rmaps-base.txt", "orte-rmaps-base:alloc-error",
-                                       true, app->num_procs, app->app);
-                        return ORTE_ERR_SILENT;
-                    } else if (ORTE_MAPPING_NO_OVERSUBSCRIBE & ORTE_GET_MAPPING_DIRECTIVE(jdata->map->mapping)) {
-                        /* if we were explicitly told not to oversubscribe, then don't */
-                        orte_show_help("help-orte-rmaps-base.txt", "orte-rmaps-base:alloc-error",
-                                       true, app->num_procs, app->app);
-                        return ORTE_ERR_SILENT;
+                                       true, node->num_procs, app->app);
+                        ORTE_UPDATE_EXIT_STATUS(ORTE_ERROR_DEFAULT_EXIT_CODE);
+                        rc = ORTE_ERR_SILENT;
+                        goto error;
+                    }
+                    /* flag the node as oversubscribed so that sched-yield gets
+                     * properly set
+                     */
+                    ORTE_FLAG_SET(node, ORTE_NODE_FLAG_OVERSUBSCRIBED);
+                    ORTE_FLAG_SET(jdata, ORTE_JOB_FLAG_OVERSUBSCRIBED);
+                    /* check for permission */
+                    if (ORTE_FLAG_TEST(node, ORTE_NODE_FLAG_SLOTS_GIVEN)) {
+                        /* if we weren't given a directive either way, then we will error out
+                         * as the #slots were specifically given, either by the host RM or
+                         * via hostfile/dash-host */
+                        if (!(ORTE_MAPPING_SUBSCRIBE_GIVEN & ORTE_GET_MAPPING_DIRECTIVE(orte_rmaps_base.mapping))) {
+                            orte_show_help("help-orte-rmaps-base.txt", "orte-rmaps-base:alloc-error",
+                                           true, app->num_procs, app->app);
+                            ORTE_UPDATE_EXIT_STATUS(ORTE_ERROR_DEFAULT_EXIT_CODE);
+                            return ORTE_ERR_SILENT;
+                        } else if (ORTE_MAPPING_NO_OVERSUBSCRIBE & ORTE_GET_MAPPING_DIRECTIVE(jdata->map->mapping)) {
+                            /* if we were explicitly told not to oversubscribe, then don't */
+                            orte_show_help("help-orte-rmaps-base.txt", "orte-rmaps-base:alloc-error",
+                                           true, app->num_procs, app->app);
+                            ORTE_UPDATE_EXIT_STATUS(ORTE_ERROR_DEFAULT_EXIT_CODE);
+                            return ORTE_ERR_SILENT;
+                        }
                     }
                 }
             }
@@ -400,11 +388,6 @@ static int ppr_mapper(orte_job_t *jdata)
             rc = ORTE_ERR_SILENT;
             goto error;
         }
-        /* compute vpids and add proc objects to the job */
-        if (ORTE_SUCCESS != (rc = orte_rmaps_base_compute_vpids(jdata, app, &node_list))) {
-            ORTE_ERROR_LOG(rc);
-            goto error;
-        }
 
         /* track the total number of processes we mapped - must update
          * this AFTER we compute vpids so that computation is done
@@ -419,7 +402,7 @@ static int ppr_mapper(orte_job_t *jdata)
     }
     return ORTE_SUCCESS;
 
- error:
+  error:
     while (NULL != (item = opal_list_remove_first(&node_list))) {
         OBJ_RELEASE(item);
     }
@@ -427,7 +410,6 @@ static int ppr_mapper(orte_job_t *jdata)
     return rc;
 }
 
-#if OPAL_HAVE_HWLOC
 static hwloc_obj_t find_split(hwloc_topology_t topo, hwloc_obj_t obj)
 {
     unsigned k;
@@ -464,6 +446,7 @@ static void prune(orte_jobid_t jobid,
     orte_proc_t *proc, *pptr, *procmax;
     opal_hwloc_level_t ll;
     char dang[64];
+    hwloc_obj_t locale;
 
     opal_output_verbose(5, orte_rmaps_base_framework.framework_output,
                         "mca:rmaps:ppr: pruning level %d",
@@ -497,7 +480,7 @@ static void prune(orte_jobid_t jobid,
     }
 
     /* get the number of resources at this level on this node */
-    nobjs = opal_hwloc_base_get_nbobjs_by_type(node->topology,
+    nobjs = opal_hwloc_base_get_nbobjs_by_type(node->topology->topo,
                                                lvl, cache_level,
                                                OPAL_HWLOC_AVAILABLE);
 
@@ -505,11 +488,11 @@ static void prune(orte_jobid_t jobid,
      * underneath it and check against the limit
      */
     for (i=0; i < nobjs; i++) {
-        obj = opal_hwloc_base_get_obj_by_type(node->topology,
+        obj = opal_hwloc_base_get_obj_by_type(node->topology->topo,
                                               lvl, cache_level,
                                               i, OPAL_HWLOC_AVAILABLE);
         /* get the available cpuset */
-        avail = opal_hwloc_base_get_available_cpus(node->topology, obj);
+        avail = opal_hwloc_base_get_available_cpus(node->topology->topo, obj);
 
         /* look at the intersection of this object's cpuset and that
          * of each proc in the job/app - if they intersect, then count this proc
@@ -524,7 +507,12 @@ static void prune(orte_jobid_t jobid,
                 proc->app_idx != app_idx) {
                 continue;
             }
-            cpus = opal_hwloc_base_get_available_cpus(node->topology, proc->locale);
+            locale = NULL;
+            if (orte_get_attribute(&proc->attributes, ORTE_PROC_HWLOC_LOCALE, (void**)&locale, OPAL_PTR)) {
+                ORTE_ERROR_LOG(ORTE_ERR_NOT_FOUND);
+                return;
+            }
+            cpus = opal_hwloc_base_get_available_cpus(node->topology->topo, locale);
             if (hwloc_bitmap_intersects(avail, cpus)) {
                 nprocs++;
             }
@@ -550,7 +538,7 @@ static void prune(orte_jobid_t jobid,
              * have only one child, then return this
              * object
              */
-            top = find_split(node->topology, obj);
+            top = find_split(node->topology->topo, obj);
             hwloc_obj_type_snprintf(dang, 64, top, 1);
             opal_output_verbose(5, orte_rmaps_base_framework.framework_output,
                                 "mca:rmaps:ppr: SPLIT AT LEVEL %s", dang);
@@ -562,7 +550,7 @@ static void prune(orte_jobid_t jobid,
             /* find the child with the most procs underneath it */
             for (k=0; k < top->arity && limit < nprocs; k++) {
                 /* get this object's available cpuset */
-                childcpus = opal_hwloc_base_get_available_cpus(node->topology, top->children[k]);
+                childcpus = opal_hwloc_base_get_available_cpus(node->topology->topo, top->children[k]);
                 nunder = 0;
                 pptr = NULL;
                 for (n=0; n < node->procs->size; n++) {
@@ -573,7 +561,12 @@ static void prune(orte_jobid_t jobid,
                         proc->app_idx != app_idx) {
                         continue;
                     }
-                    cpus = opal_hwloc_base_get_available_cpus(node->topology, proc->locale);
+                    locale = NULL;
+                    if (orte_get_attribute(&proc->attributes, ORTE_PROC_HWLOC_LOCALE, (void**)&locale, OPAL_PTR)) {
+                        ORTE_ERROR_LOG(ORTE_ERR_NOT_FOUND);
+                        return;
+                    }
+                    cpus = opal_hwloc_base_get_available_cpus(node->topology->topo, locale);
                     if (hwloc_bitmap_intersects(childcpus, cpus)) {
                         nunder++;
                         if (NULL == pptr) {
@@ -622,4 +615,122 @@ static void prune(orte_jobid_t jobid,
  error:
     opal_output(0, "INFINITE LOOP");
 }
-#endif
+
+static int assign_locations(orte_job_t *jdata)
+{
+    int i, j, m, n;
+    mca_base_component_t *c=&mca_rmaps_ppr_component.base_version;
+    orte_node_t *node;
+    orte_proc_t *proc;
+    orte_app_context_t *app;
+    opal_hwloc_level_t level;
+    hwloc_obj_t obj;
+    unsigned int cache_level=0;
+    int ppr, cnt, nobjs, nprocs_mapped;
+    char **ppr_req, **ck;
+
+    if (NULL == jdata->map->last_mapper ||
+        0 != strcasecmp(jdata->map->last_mapper, c->mca_component_name)) {
+        /* a mapper has been specified, and it isn't me */
+        opal_output_verbose(5, orte_rmaps_base_framework.framework_output,
+                            "mca:rmaps:ppr: job %s not using ppr assign: %s",
+                            ORTE_JOBID_PRINT(jdata->jobid),
+                            (NULL == jdata->map->last_mapper) ? "NULL" : jdata->map->last_mapper);
+        return ORTE_ERR_TAKE_NEXT_OPTION;
+    }
+
+    opal_output_verbose(5, orte_rmaps_base_framework.framework_output,
+                        "mca:rmaps:ppr: assigning locations for job %s with ppr %s policy %s",
+                        ORTE_JOBID_PRINT(jdata->jobid), jdata->map->ppr,
+                        orte_rmaps_base_print_mapping(jdata->map->mapping));
+
+    /* pickup the object level */
+    if (ORTE_MAPPING_BYNODE == ORTE_GET_MAPPING_POLICY(jdata->map->mapping)) {
+        level = OPAL_HWLOC_NODE_LEVEL;
+    } else if (ORTE_MAPPING_BYHWTHREAD == ORTE_GET_MAPPING_POLICY(jdata->map->mapping)) {
+        level = OPAL_HWLOC_HWTHREAD_LEVEL;
+    } else if (ORTE_MAPPING_BYCORE == ORTE_GET_MAPPING_POLICY(jdata->map->mapping)) {
+        level = OPAL_HWLOC_CORE_LEVEL;
+    } else if (ORTE_MAPPING_BYSOCKET == ORTE_GET_MAPPING_POLICY(jdata->map->mapping)) {
+        level = OPAL_HWLOC_SOCKET_LEVEL;
+    } else if (ORTE_MAPPING_BYL1CACHE == ORTE_GET_MAPPING_POLICY(jdata->map->mapping)) {
+        level = OPAL_HWLOC_L1CACHE_LEVEL;
+        cache_level = 1;
+    } else if (ORTE_MAPPING_BYL2CACHE == ORTE_GET_MAPPING_POLICY(jdata->map->mapping)) {
+        level = OPAL_HWLOC_L2CACHE_LEVEL;
+        cache_level = 2;
+    } else if (ORTE_MAPPING_BYL3CACHE == ORTE_GET_MAPPING_POLICY(jdata->map->mapping)) {
+        level = OPAL_HWLOC_L3CACHE_LEVEL;
+        cache_level = 3;
+    } else if (ORTE_MAPPING_BYNUMA == ORTE_GET_MAPPING_POLICY(jdata->map->mapping)) {
+        level = OPAL_HWLOC_NUMA_LEVEL;
+    } else {
+        ORTE_ERROR_LOG(ORTE_ERR_NOT_FOUND);
+        return ORTE_ERR_TAKE_NEXT_OPTION;
+    }
+
+    /* get the ppr value */
+    ppr_req = opal_argv_split(jdata->map->ppr, ',');
+    ck = opal_argv_split(ppr_req[0], ':');
+    ppr = strtol(ck[0], NULL, 10);
+    opal_argv_free(ck);
+    opal_argv_free(ppr_req);
+
+    /* start assigning procs to objects, filling each object as we go until
+     * all procs are assigned. */
+    for (n=0; n < jdata->apps->size; n++) {
+        if (NULL == (app = (orte_app_context_t*)opal_pointer_array_get_item(jdata->apps, n))) {
+            continue;
+        }
+        nprocs_mapped = 0;
+        for (m=0; m < jdata->map->nodes->size; m++) {
+            if (NULL == (node = (orte_node_t*)opal_pointer_array_get_item(jdata->map->nodes, m))) {
+                continue;
+            }
+            if (NULL == node->topology || NULL == node->topology->topo) {
+                orte_show_help("help-orte-rmaps-ppr.txt", "ppr-topo-missing",
+                               true, node->name);
+                return ORTE_ERR_SILENT;
+            }
+            if (OPAL_HWLOC_NODE_LEVEL == level) {
+                obj = hwloc_get_root_obj(node->topology->topo);
+                for (j=0; j < node->procs->size; j++) {
+                    if (NULL == (proc = (orte_proc_t*)opal_pointer_array_get_item(node->procs, j))) {
+                        continue;
+                    }
+                    if (proc->name.jobid != jdata->jobid) {
+                        continue;
+                    }
+                    orte_set_attribute(&proc->attributes, ORTE_PROC_HWLOC_LOCALE, ORTE_ATTR_LOCAL, obj, OPAL_PTR);
+                }
+            } else {
+                /* get the number of resources on this node at this level */
+                nobjs = opal_hwloc_base_get_nbobjs_by_type(node->topology->topo,
+                                                           level, cache_level,
+                                                           OPAL_HWLOC_AVAILABLE);
+
+                /* map the specified number of procs to each such resource on this node,
+                 * recording the locale of each proc so we know its cpuset
+                 */
+                cnt = 0;
+                for (i=0; i < nobjs; i++) {
+                    obj = opal_hwloc_base_get_obj_by_type(node->topology->topo,
+                                                          level, cache_level,
+                                                          i, OPAL_HWLOC_AVAILABLE);
+                    for (j=0; j < node->procs->size && cnt < ppr && nprocs_mapped < app->num_procs; j++) {
+                        if (NULL == (proc = (orte_proc_t*)opal_pointer_array_get_item(node->procs, j))) {
+                            continue;
+                        }
+                        if (proc->name.jobid != jdata->jobid) {
+                            continue;
+                        }
+                        nprocs_mapped++;
+                        cnt++;
+                        orte_set_attribute(&proc->attributes, ORTE_PROC_HWLOC_LOCALE, ORTE_ATTR_LOCAL, obj, OPAL_PTR);
+                    }
+                }
+            }
+        }
+    }
+    return ORTE_SUCCESS;
+}
