@@ -13,10 +13,10 @@
  *                         All rights reserved.
  * Copyright (c) 2009      Cisco Systems, Inc.  All rights reserved.
  * Copyright (c) 2011      Oak Ridge National Labs.  All rights reserved.
- * Copyright (c) 2013-2015 Intel, Inc.  All rights reserved.
+ * Copyright (c) 2013-2017 Intel, Inc. All rights reserved.
  * Copyright (c) 2014      Mellanox Technologies, Inc.
  *                         All rights reserved.
- * Copyright (c) 2014      Research Organization for Information Science
+ * Copyright (c) 2014-2016 Research Organization for Information Science
  *                         and Technology (RIST). All rights reserved.
  * $COPYRIGHT$
  *
@@ -38,21 +38,134 @@
 
 #include "orte/mca/errmgr/errmgr.h"
 #include "orte/util/name_fns.h"
+#include "orte/util/show_help.h"
+#include "orte/util/threads.h"
 #include "orte/runtime/orte_data_server.h"
 #include "orte/runtime/orte_globals.h"
 #include "orte/mca/rml/rml.h"
+#include "orte/mca/rml/base/rml_contact.h"
 
 #include "pmix_server_internal.h"
 
+static int init_server(void)
+{
+    char *server;
+    opal_buffer_t buf;
+    char input[1024], *filename;
+    FILE *fp;
+    int rc;
+
+    /* only do this once */
+    orte_pmix_server_globals.pubsub_init = true;
+
+    /* if the universal server wasn't specified, then we use
+     * our own HNP for that purpose */
+    if (NULL == orte_data_server_uri) {
+        orte_pmix_server_globals.server = *ORTE_PROC_MY_HNP;
+    } else {
+        if (0 == strncmp(orte_data_server_uri, "file", strlen("file")) ||
+            0 == strncmp(orte_data_server_uri, "FILE", strlen("FILE"))) {
+            /* it is a file - get the filename */
+            filename = strchr(orte_data_server_uri, ':');
+            if (NULL == filename) {
+                /* filename is not correctly formatted */
+                orte_show_help("help-orterun.txt", "orterun:ompi-server-filename-bad", true,
+                               orte_basename, orte_data_server_uri);
+                return ORTE_ERR_BAD_PARAM;
+            }
+            ++filename; /* space past the : */
+
+            if (0 >= strlen(filename)) {
+                /* they forgot to give us the name! */
+                orte_show_help("help-orterun.txt", "orterun:ompi-server-filename-missing", true,
+                               orte_basename, orte_data_server_uri);
+                return ORTE_ERR_BAD_PARAM;
+            }
+
+            /* open the file and extract the uri */
+            fp = fopen(filename, "r");
+            if (NULL == fp) { /* can't find or read file! */
+                orte_show_help("help-orterun.txt", "orterun:ompi-server-filename-access", true,
+                               orte_basename, orte_data_server_uri);
+                return ORTE_ERR_BAD_PARAM;
+            }
+            if (NULL == fgets(input, 1024, fp)) {
+                /* something malformed about file */
+                fclose(fp);
+                orte_show_help("help-orterun.txt", "orterun:ompi-server-file-bad", true,
+                               orte_basename, orte_data_server_uri,
+                               orte_basename);
+                return ORTE_ERR_BAD_PARAM;
+            }
+            fclose(fp);
+            input[strlen(input)-1] = '\0';  /* remove newline */
+            server = strdup(input);
+        } else {
+            server = strdup(orte_data_server_uri);
+        }
+        /* setup our route to the server */
+        OBJ_CONSTRUCT(&buf, opal_buffer_t);
+        opal_dss.pack(&buf, &server, 1, OPAL_STRING);
+        if (ORTE_SUCCESS != (rc = orte_rml_base_update_contact_info(&buf))) {
+            ORTE_ERROR_LOG(rc);
+            ORTE_UPDATE_EXIT_STATUS(ORTE_ERROR_DEFAULT_EXIT_CODE);
+            return rc;
+        }
+        OBJ_DESTRUCT(&buf);
+        /* parse the URI to get the server's name */
+        if (ORTE_SUCCESS != (rc = orte_rml_base_parse_uris(server, &orte_pmix_server_globals.server, NULL))) {
+            ORTE_ERROR_LOG(rc);
+            return rc;
+        }
+        /* check if we are to wait for the server to start - resolves
+         * a race condition that can occur when the server is run
+         * as a background job - e.g., in scripts
+         */
+        if (orte_pmix_server_globals.wait_for_server) {
+            /* ping the server */
+            struct timeval timeout;
+            timeout.tv_sec = orte_pmix_server_globals.timeout;
+            timeout.tv_usec = 0;
+            if (ORTE_SUCCESS != (rc = orte_rml.ping(orte_mgmt_conduit, server, &timeout))) {
+                /* try it one more time */
+                if (ORTE_SUCCESS != (rc = orte_rml.ping(orte_mgmt_conduit, server, &timeout))) {
+                    /* okay give up */
+                    orte_show_help("help-orterun.txt", "orterun:server-not-found", true,
+                                   orte_basename, server,
+                                   (long)orte_pmix_server_globals.timeout,
+                                   ORTE_ERROR_NAME(rc));
+                    ORTE_UPDATE_EXIT_STATUS(ORTE_ERROR_DEFAULT_EXIT_CODE);
+                    return rc;
+                }
+            }
+        }
+    }
+
+    return ORTE_SUCCESS;
+}
+
 static void execute(int sd, short args, void *cbdata)
 {
     pmix_server_req_t *req = (pmix_server_req_t*)cbdata;
     int rc;
     opal_buffer_t *xfer;
+    orte_process_name_t *target;
+
+    ORTE_ACQUIRE_OBJECT(req);
+
+    if (!orte_pmix_server_globals.pubsub_init) {
+        /* we need to initialize our connection to the server */
+        if (ORTE_SUCCESS != (rc = init_server())) {
+            orte_show_help("help-orted.txt", "noserver", true,
+                           (NULL == orte_data_server_uri) ?
+                           "NULL" : orte_data_server_uri);
+            goto callback;
+        }
+    }
 
     /* add this request to our tracker hotel */
     if (OPAL_SUCCESS != (rc = opal_hotel_checkin(&orte_pmix_server_globals.reqs, req, &req->room_num))) {
-        ORTE_ERROR_LOG(rc);
+        orte_show_help("help-orted.txt", "noroom", true, req->operation, orte_pmix_server_globals.num_rooms);
         goto callback;
     }
 
@@ -66,8 +179,16 @@ static void execute(int sd, short args, void *cbdata)
     }
     opal_dss.copy_payload(xfer, &req->msg);
 
+    /* if the range is SESSION, then set the target to the global server */
+    if (OPAL_PMIX_RANGE_SESSION == req->range) {
+        target = &orte_pmix_server_globals.server;
+    } else {
+        target = ORTE_PROC_MY_HNP;
+    }
+
     /* send the request to the target */
-    rc = orte_rml.send_buffer_nb(&req->target, xfer,
+    rc = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                 target, xfer,
                                  ORTE_RML_TAG_DATA_SERVER,
                                  orte_rml_send_callback, NULL);
     if (ORTE_SUCCESS == rc) {
@@ -93,12 +214,16 @@ int pmix_server_publish_fn(opal_process_name_t *proc,
     int rc;
     uint8_t cmd = ORTE_PMIX_PUBLISH_CMD;
     opal_value_t *iptr;
-    opal_pmix_data_range_t range = OPAL_PMIX_SESSION;
     opal_pmix_persistence_t persist = OPAL_PMIX_PERSIST_APP;
     bool rset, pset;
 
+    opal_output_verbose(1, orte_pmix_server_globals.output,
+                        "%s orted:pmix:server PUBLISH",
+                        ORTE_NAME_PRINT(ORTE_PROC_MY_NAME));
+
     /* create the caddy */
     req = OBJ_NEW(pmix_server_req_t);
+    (void)asprintf(&req->operation, "PUBLISH: %s:%d", __FILE__, __LINE__);
     req->opcbfunc = cbfunc;
     req->cbdata = cbdata;
 
@@ -121,13 +246,13 @@ int pmix_server_publish_fn(opal_process_name_t *proc,
     pset = false;
     OPAL_LIST_FOREACH(iptr, info, opal_value_t) {
         if (0 == strcmp(iptr->key, OPAL_PMIX_RANGE)) {
-            range = iptr->data.integer;
+            req->range = (opal_pmix_data_range_t)iptr->data.uint;
             if (pset) {
                 break;
             }
             rset = true;
         } else if (0 == strcmp(iptr->key, OPAL_PMIX_PERSISTENCE)) {
-            persist = iptr->data.integer;
+            persist = (opal_pmix_persistence_t)iptr->data.integer;
             if (rset) {
                 break;
             }
@@ -136,19 +261,12 @@ int pmix_server_publish_fn(opal_process_name_t *proc,
     }
 
     /* pack the range */
-    if (OPAL_SUCCESS != (rc = opal_dss.pack(&req->msg, &range, 1, OPAL_INT))) {
+    if (OPAL_SUCCESS != (rc = opal_dss.pack(&req->msg, &req->range, 1, OPAL_PMIX_DATA_RANGE))) {
         ORTE_ERROR_LOG(rc);
         OBJ_RELEASE(req);
         return rc;
     }
 
-    /* if the range is SESSION, then set the target to the global server */
-    if (OPAL_PMIX_SESSION == range) {
-        req->target = orte_pmix_server_globals.server;
-    } else {
-        req->target = *ORTE_PROC_MY_HNP;
-    }
-
     /* pack the persistence */
     if (OPAL_SUCCESS != (rc = opal_dss.pack(&req->msg, &persist, 1, OPAL_INT))) {
         ORTE_ERROR_LOG(rc);
@@ -156,13 +274,18 @@ int pmix_server_publish_fn(opal_process_name_t *proc,
         return rc;
     }
 
-    /* if we have items, pack those too - ignore persistence
+    /* if we have items, pack those too - ignore persistence, timeout
      * and range values */
     OPAL_LIST_FOREACH(iptr, info, opal_value_t) {
         if (0 == strcmp(iptr->key, OPAL_PMIX_RANGE) ||
             0 == strcmp(iptr->key, OPAL_PMIX_PERSISTENCE)) {
             continue;
         }
+        if (0 == strcmp(iptr->key, OPAL_PMIX_TIMEOUT)) {
+            /* record the timeout value, but don't pack it */
+            req->timeout = iptr->data.integer;
+            continue;
+        }
         opal_output_verbose(5, orte_pmix_server_globals.output,
                             "%s publishing data %s of type %d from source %s",
                             ORTE_NAME_PRINT(ORTE_PROC_MY_NAME), iptr->key, iptr->type,
@@ -178,6 +301,7 @@ int pmix_server_publish_fn(opal_process_name_t *proc,
     opal_event_set(orte_event_base, &(req->ev),
                    -1, OPAL_EV_WRITE, execute, req);
     opal_event_set_priority(&(req->ev), ORTE_MSG_PRI);
+    ORTE_POST_OBJECT(req);
     opal_event_active(&(req->ev), OPAL_EV_WRITE, 1);
 
     return OPAL_SUCCESS;
@@ -193,7 +317,6 @@ int pmix_server_lookup_fn(opal_process_name_t *proc, char **keys,
     uint8_t cmd = ORTE_PMIX_LOOKUP_CMD;
     int32_t nkeys, i;
     opal_value_t *iptr;
-    opal_pmix_data_range_t range = OPAL_PMIX_SESSION;
 
     /* the list of info objects are directives for us - they include
      * things like timeout constraints, so there is no reason to
@@ -201,6 +324,7 @@ int pmix_server_lookup_fn(opal_process_name_t *proc, char **keys,
 
     /* create the caddy */
     req = OBJ_NEW(pmix_server_req_t);
+    (void)asprintf(&req->operation, "LOOKUP: %s:%d", __FILE__, __LINE__);
     req->lkcbfunc = cbfunc;
     req->cbdata = cbdata;
 
@@ -211,26 +335,26 @@ int pmix_server_lookup_fn(opal_process_name_t *proc, char **keys,
         return rc;
     }
 
-    /* no help for it - need to search for range */
-    OPAL_LIST_FOREACH(iptr, info, opal_value_t) {
-        if (0 == strcmp(iptr->key, OPAL_PMIX_RANGE)) {
-            range = iptr->data.integer;
-            break;
-        }
-    }
-
-    /* pack the range */
-    if (OPAL_SUCCESS != (rc = opal_dss.pack(&req->msg, &range, 1, OPAL_INT))) {
+    /* pack the requesting process jobid */
+    if (OPAL_SUCCESS != (rc = opal_dss.pack(&req->msg, &proc->jobid, 1, ORTE_JOBID))) {
         ORTE_ERROR_LOG(rc);
         OBJ_RELEASE(req);
         return rc;
     }
 
-    /* if the range is SESSION, then set the target to the global server */
-    if (OPAL_PMIX_SESSION == range) {
-        req->target = orte_pmix_server_globals.server;
-    } else {
-        req->target = *ORTE_PROC_MY_HNP;
+    /* no help for it - need to search for range */
+    OPAL_LIST_FOREACH(iptr, info, opal_value_t) {
+        if (0 == strcmp(iptr->key, OPAL_PMIX_RANGE)) {
+            req->range = (opal_pmix_data_range_t)iptr->data.uint;
+            break;
+        }
+    }
+
+    /* pack the range */
+    if (OPAL_SUCCESS != (rc = opal_dss.pack(&req->msg, &req->range, 1, OPAL_PMIX_DATA_RANGE))) {
+        ORTE_ERROR_LOG(rc);
+        OBJ_RELEASE(req);
+        return rc;
     }
 
     /* pack the number of keys */
@@ -243,6 +367,10 @@ int pmix_server_lookup_fn(opal_process_name_t *proc, char **keys,
 
     /* pack the keys too */
     for (i=0; i < nkeys; i++) {
+        opal_output_verbose(5, orte_pmix_server_globals.output,
+                            "%s lookup data %s for proc %s",
+                            ORTE_NAME_PRINT(ORTE_PROC_MY_NAME), keys[i],
+                            ORTE_NAME_PRINT(proc));
         if (OPAL_SUCCESS != (rc = opal_dss.pack(&req->msg, &keys[i], 1, OPAL_STRING))) {
             ORTE_ERROR_LOG(rc);
             OBJ_RELEASE(req);
@@ -250,11 +378,16 @@ int pmix_server_lookup_fn(opal_process_name_t *proc, char **keys,
         }
     }
 
-    /* if we have items, pack those too - ignore range value */
+    /* if we have items, pack those too - ignore range and timeout value */
     OPAL_LIST_FOREACH(iptr, info, opal_value_t) {
         if (0 == strcmp(iptr->key, OPAL_PMIX_RANGE)) {
             continue;
         }
+        if (0 == strcmp(iptr->key, OPAL_PMIX_TIMEOUT)) {
+            /* record the timeout value, but don't pack it */
+            req->timeout = iptr->data.integer;
+            continue;
+        }
         if (OPAL_SUCCESS != (rc = opal_dss.pack(&req->msg, &iptr, 1, OPAL_VALUE))) {
             ORTE_ERROR_LOG(rc);
             OBJ_RELEASE(req);
@@ -266,6 +399,7 @@ int pmix_server_lookup_fn(opal_process_name_t *proc, char **keys,
     opal_event_set(orte_event_base, &(req->ev),
                    -1, OPAL_EV_WRITE, execute, req);
     opal_event_set_priority(&(req->ev), ORTE_MSG_PRI);
+    ORTE_POST_OBJECT(req);
     opal_event_active(&(req->ev), OPAL_EV_WRITE, 1);
 
     return OPAL_SUCCESS;
@@ -280,10 +414,10 @@ int pmix_server_unpublish_fn(opal_process_name_t *proc, char **keys,
     uint8_t cmd = ORTE_PMIX_UNPUBLISH_CMD;
     uint32_t nkeys, n;
     opal_value_t *iptr;
-    opal_pmix_data_range_t range = OPAL_PMIX_SESSION;
 
     /* create the caddy */
     req = OBJ_NEW(pmix_server_req_t);
+    (void)asprintf(&req->operation, "UNPUBLISH: %s:%d", __FILE__, __LINE__);
     req->opcbfunc = cbfunc;
     req->cbdata = cbdata;
 
@@ -304,25 +438,18 @@ int pmix_server_unpublish_fn(opal_process_name_t *proc, char **keys,
     /* no help for it - need to search for range */
     OPAL_LIST_FOREACH(iptr, info, opal_value_t) {
         if (0 == strcmp(iptr->key, OPAL_PMIX_RANGE)) {
-            range = iptr->data.integer;
+            req->range = (opal_pmix_data_range_t)iptr->data.integer;
             break;
         }
     }
 
     /* pack the range */
-    if (OPAL_SUCCESS != (rc = opal_dss.pack(&req->msg, &range, 1, OPAL_INT))) {
+    if (OPAL_SUCCESS != (rc = opal_dss.pack(&req->msg, &req->range, 1, OPAL_INT))) {
         ORTE_ERROR_LOG(rc);
         OBJ_RELEASE(req);
         return rc;
     }
 
-    /* if the range is SESSION, then set the target to the global server */
-    if (OPAL_PMIX_SESSION == range) {
-        req->target = orte_pmix_server_globals.server;
-    } else {
-        req->target = *ORTE_PROC_MY_HNP;
-    }
-
     /* pack the number of keys */
     nkeys = opal_argv_count(keys);
     if (OPAL_SUCCESS != (rc = opal_dss.pack(&req->msg, &nkeys, 1, OPAL_UINT32))) {
@@ -340,11 +467,16 @@ int pmix_server_unpublish_fn(opal_process_name_t *proc, char **keys,
         }
     }
 
-    /* if we have items, pack those too - ignore range value */
+    /* if we have items, pack those too - ignore range and timeout value */
     OPAL_LIST_FOREACH(iptr, info, opal_value_t) {
         if (0 == strcmp(iptr->key, OPAL_PMIX_RANGE)) {
             continue;
         }
+        if (0 == strcmp(iptr->key, OPAL_PMIX_TIMEOUT)) {
+            /* record the timeout value, but don't pack it */
+            req->timeout = iptr->data.integer;
+            continue;
+        }
         if (OPAL_SUCCESS != (rc = opal_dss.pack(&req->msg, &iptr, 1, OPAL_VALUE))) {
             ORTE_ERROR_LOG(rc);
             OBJ_RELEASE(req);
@@ -356,6 +488,7 @@ int pmix_server_unpublish_fn(opal_process_name_t *proc, char **keys,
     opal_event_set(orte_event_base, &(req->ev),
                    -1, OPAL_EV_WRITE, execute, req);
     opal_event_set_priority(&(req->ev), ORTE_MSG_PRI);
+    ORTE_POST_OBJECT(req);
     opal_event_active(&(req->ev), OPAL_EV_WRITE, 1);
 
     return OPAL_SUCCESS;
@@ -445,4 +578,3 @@ void pmix_server_keyval_client(int status, orte_process_name_t* sender,
         OBJ_RELEASE(req);
     }
 }
-
