@@ -2,17 +2,20 @@
  * Copyright (c) 2004-2005 The Trustees of Indiana University and Indiana
  *                         University Research and Technology
  *                         Corporation.  All rights reserved.
- * Copyright (c) 2004-2006 The University of Tennessee and The University
+ * Copyright (c) 2004-2011 The University of Tennessee and The University
  *                         of Tennessee Research Foundation.  All rights
  *                         reserved.
- * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart, 
+ * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart,
  *                         University of Stuttgart.  All rights reserved.
  * Copyright (c) 2004-2005 The Regents of the University of California.
  *                         All rights reserved.
+ * Copyright (c) 2010-2012 Los Alamos National Security, LLC.
+ *                         All rights reserved.
+ * Copyright (c) 2014-2017 Intel, Inc. All rights reserved.
  * $COPYRIGHT$
- * 
+ *
  * Additional copyrights may follow
- * 
+ *
  * $HEADER$
  */
 #include "orte_config.h"
@@ -24,6 +27,8 @@
 
 #include "opal/util/output.h"
 #include "opal/threads/tsd.h"
+#include "opal/mca/event/event.h"
+#include "opal/runtime/opal_progress.h"
 
 #include "opal/dss/dss.h"
 #include "orte/mca/errmgr/errmgr.h"
@@ -33,6 +38,7 @@
 #include "orte/mca/rml/base/rml_contact.h"
 #include "orte/mca/routed/routed.h"
 #include "orte/util/name_fns.h"
+#include "orte/util/threads.h"
 #include "orte/runtime/orte_globals.h"
 #include "orte/runtime/orte_wait.h"
 
@@ -47,10 +53,15 @@ static int error_exit;
 
 static void quicktime_cb(int fd, short event, void *cbdata)
 {
+    /* release the timer */
     if (NULL != quicktime) {
-        free(quicktime);
+        opal_event_free(quicktime);
         quicktime = NULL;
     }
+
+    /* cancel the recv */
+    orte_rml.recv_cancel(ORTE_NAME_WILDCARD, ORTE_RML_TAG_TOOL);
+
     error_exit = ORTE_ERR_SILENT;
     /* declare it fired */
     timer_fired = true;
@@ -62,12 +73,13 @@ static void send_cbfunc(int status, orte_process_name_t* sender,
 {
     /* cancel the timer */
     if (NULL != quicktime) {
-        opal_evtimer_del(quicktime);
-        free(quicktime);
+        opal_event_free(quicktime);
         quicktime = NULL;
     }
     /* declare the work done */
     timer_fired = true;
+    /* release the message */
+    OBJ_RELEASE(buffer);
 }
 
 static void recv_info(int status, orte_process_name_t* sender,
@@ -75,11 +87,10 @@ static void recv_info(int status, orte_process_name_t* sender,
                       void* cbdata)
 {
     int rc;
-    
+
     /* cancel the timer */
     if (NULL != quicktime) {
-        opal_evtimer_del(quicktime);
-        free(quicktime);
+        opal_event_free (quicktime);
         quicktime = NULL;
     }
     /* xfer the answer */
@@ -99,25 +110,22 @@ static bool tool_connected = false;
 int orte_util_comm_connect_tool(char *uri)
 {
     int rc;
-    
+
     /* set the contact info into the comm hash tables*/
-    if (ORTE_SUCCESS != (rc = orte_rml.set_contact_info(uri))) {
-        ORTE_ERROR_LOG(rc);
-        return(rc);
-    }
-    
+    orte_rml.set_contact_info(uri);
+
     /* extract the tool's name and store it */
     if (ORTE_SUCCESS != (rc = orte_rml_base_parse_uris(uri, &tool, NULL))) {
         ORTE_ERROR_LOG(rc);
         return rc;
     }
-    
+
     /* set the route to be direct */
-    if (ORTE_SUCCESS != (rc = orte_routed.update_route(&tool, &tool))) {
+    if (ORTE_SUCCESS != (rc = orte_routed.update_route(NULL, &tool, &tool))) {
         ORTE_ERROR_LOG(rc);
         return rc;
     }
-    
+
     tool_connected = true;
     return ORTE_SUCCESS;
 }
@@ -129,18 +137,19 @@ static bool step=false;
 int orte_util_comm_report_event(orte_comm_event_t ev)
 {
     int rc, i;
-    opal_buffer_t buf;
+    opal_buffer_t *buf;
     orte_node_t *node;
-    
+    struct timeval tv;
+
     /* if nothing is connected, ignore this */
     if (!tool_connected) {
         return ORTE_SUCCESS;
     }
-    
+
     /* init a buffer for the data */
-    OBJ_CONSTRUCT(&buf, opal_buffer_t);
+    buf = OBJ_NEW(opal_buffer_t);
     /* flag the type of event */
-    opal_dss.pack(&buf, &ev, 1, ORTE_COMM_EVENT);
+    opal_dss.pack(buf, &ev, 1, ORTE_COMM_EVENT);
 
     switch (ev) {
         case ORTE_COMM_EVENT_ALLOCATE:
@@ -149,30 +158,49 @@ int orte_util_comm_report_event(orte_comm_event_t ev)
                 if (NULL == (node = (orte_node_t*)opal_pointer_array_get_item(orte_node_pool, i))) {
                     continue;
                 }
-                opal_dss.pack(&buf, &node->name, 1, OPAL_STRING);
+                opal_dss.pack(buf, &node->name, 1, OPAL_STRING);
             }
             break;
-        
+
         case ORTE_COMM_EVENT_MAP:
             break;
-        
+
         case ORTE_COMM_EVENT_LAUNCH:
             break;
-        
+
         default:
             ORTE_ERROR_LOG(ORTE_ERROR);
-            OBJ_DESTRUCT(&buf);
+            OBJ_RELEASE(buf);
             return ORTE_ERROR;
             break;
     }
-    
+
+   /* define a max time to wait for send to complete */
+    timer_fired = false;
+    error_exit = ORTE_SUCCESS;
+    quicktime = opal_event_alloc();
+    tv.tv_sec = 0;
+    tv.tv_usec = 100000;
+    opal_event_evtimer_set(orte_event_base, quicktime, quicktime_cb, NULL);
+    opal_event_set_priority(quicktime, ORTE_ERROR_PRI);
+    opal_event_evtimer_add(quicktime, &tv);
+
     /* do the send */
-    if (0 > (rc = orte_rml.send_buffer(&tool, &buf, ORTE_RML_TAG_TOOL, 0))) {
+    if (0 > (rc = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                          &tool, buf, ORTE_RML_TAG_TOOL, send_cbfunc, NULL))) {
         ORTE_ERROR_LOG(rc);
-        OBJ_DESTRUCT(&buf);
+        OBJ_RELEASE(buf);
         return rc;
     }
-    
+
+    while (!timer_fired) {
+        opal_progress();
+    }
+
+    if (ORTE_SUCCESS != error_exit) {
+        return error_exit;
+    }
+
     if (step) {
         /* the caller wants to wait until an ack is received -
          * define a max time to wait for an answer
@@ -180,27 +208,25 @@ int orte_util_comm_report_event(orte_comm_event_t ev)
         OBJ_CONSTRUCT(&answer, opal_buffer_t);
         timer_fired = false;
         error_exit = ORTE_SUCCESS;
-        ORTE_DETECT_TIMEOUT(&quicktime, 100, 1000, 100000, quicktime_cb);
-        
+
         /* get the answer */
-        if (ORTE_SUCCESS != (rc = orte_rml.recv_buffer_nb(ORTE_NAME_WILDCARD,
-                                                           ORTE_RML_TAG_TOOL,
-                                                           ORTE_RML_NON_PERSISTENT,
-                                                           recv_info,
-                                                           NULL))) {
-            /* cancel the timer */
-            if (NULL != quicktime) {
-                opal_evtimer_del(quicktime);
-                free(quicktime);
-                quicktime = NULL;
-            }
-            ORTE_ERROR_LOG(rc);
-            OBJ_DESTRUCT(&answer);
-            return rc;
+        orte_rml.recv_buffer_nb(ORTE_NAME_WILDCARD,
+                                ORTE_RML_TAG_TOOL,
+                                ORTE_RML_NON_PERSISTENT,
+                                recv_info,
+                                NULL);
+        /* set a timer for getting the answer */
+        quicktime = opal_event_alloc();
+        tv.tv_sec = 0;
+        tv.tv_usec = 100000;
+        opal_event_evtimer_set(orte_event_base, quicktime, quicktime_cb, NULL);
+        opal_event_set_priority(quicktime, ORTE_ERROR_PRI);
+        opal_event_evtimer_add(quicktime, &tv);
+
+        while (!timer_fired) {
+            opal_progress();
         }
-        
-        ORTE_PROGRESSED_WAIT(timer_fired, 0, 1);
-        
+
         /* cleanup */
         OBJ_DESTRUCT(&answer);
 
@@ -208,7 +234,7 @@ int orte_util_comm_report_event(orte_comm_event_t ev)
             return error_exit;
         }
     }
-    
+
     return ORTE_SUCCESS;
 }
 
@@ -221,11 +247,12 @@ int orte_util_comm_query_job_info(const orte_process_name_t *hnp, orte_jobid_t j
     opal_buffer_t *cmd;
     orte_daemon_cmd_flag_t command = ORTE_DAEMON_REPORT_JOB_INFO_CMD;
     orte_job_t **job_info;
+    struct timeval tv;
 
     /* set default response */
     *num_jobs = 0;
     *job_info_array = NULL;
-    
+
     /* send query to HNP */
     cmd = OBJ_NEW(opal_buffer_t);
     if (ORTE_SUCCESS != (ret = opal_dss.pack(cmd, &command, 1, ORTE_DAEMON_CMD))) {
@@ -238,69 +265,68 @@ int orte_util_comm_query_job_info(const orte_process_name_t *hnp, orte_jobid_t j
         OBJ_RELEASE(cmd);
         return ret;
     }
-    /* define a max time to wait for send to complete */
+
+   /* define a max time to wait for send to complete */
     timer_fired = false;
     error_exit = ORTE_SUCCESS;
-    ORTE_DETECT_TIMEOUT(&quicktime, 100, 1000, 100000, quicktime_cb);
-    
+    quicktime = opal_event_alloc();
+    tv.tv_sec = 0;
+    tv.tv_usec = 100000;
+    opal_event_evtimer_set(orte_event_base, quicktime, quicktime_cb, NULL);
+    opal_event_set_priority(quicktime, ORTE_ERROR_PRI);
+    opal_event_evtimer_add(quicktime, &tv);
+
     /* do the send */
-    if (0 > (ret = orte_rml.send_buffer_nb((orte_process_name_t*)hnp, cmd, ORTE_RML_TAG_DAEMON, 0,
-                                           send_cbfunc, NULL))) {
+    if (0 > (ret = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                           (orte_process_name_t*)hnp, cmd,
+                                           ORTE_RML_TAG_DAEMON, send_cbfunc, NULL))) {
         ORTE_ERROR_LOG(ret);
         OBJ_RELEASE(cmd);
         return ret;
     }
-    
-    /* wait for send to complete */
-    ORTE_PROGRESSED_WAIT(timer_fired, 0, 1);
 
-    /* release the buffer */
-    OBJ_RELEASE(cmd);
-    
-    /* did it succeed? */
-    if (ORTE_SUCCESS != error_exit) {
-        return error_exit;
+    while (!timer_fired) {
+        opal_progress();
     }
-    
+
     /* setup for answer */
     OBJ_CONSTRUCT(&answer, opal_buffer_t);
 
     /* define a max time to wait for an answer */
     timer_fired = false;
     error_exit = ORTE_SUCCESS;
-    ORTE_DETECT_TIMEOUT(&quicktime, 100, 1000, 100000, quicktime_cb);
-    
+
     /* get the answer */
-    if (ORTE_SUCCESS != (ret = orte_rml.recv_buffer_nb(ORTE_NAME_WILDCARD,
-                                                      ORTE_RML_TAG_TOOL,
-                                                      ORTE_RML_NON_PERSISTENT,
-                                                      recv_info,
-                                                      NULL))) {
-        /* cancel the timer */
-        if (NULL != quicktime) {
-            opal_evtimer_del(quicktime);
-            free(quicktime);
-            quicktime = NULL;
-        }
-        ORTE_ERROR_LOG(ret);
-        OBJ_DESTRUCT(&answer);
-        return ret;
+    orte_rml.recv_buffer_nb(ORTE_NAME_WILDCARD,
+                            ORTE_RML_TAG_TOOL,
+                            ORTE_RML_NON_PERSISTENT,
+                            recv_info,
+                            NULL);
+
+    /* set a timer for getting the answer */
+    quicktime = opal_event_alloc();
+    tv.tv_sec = 0;
+    tv.tv_usec = 100000;
+    opal_event_evtimer_set(orte_event_base, quicktime, quicktime_cb, NULL);
+    opal_event_set_priority(quicktime, ORTE_ERROR_PRI);
+    opal_event_evtimer_add(quicktime, &tv);
+
+    while (!timer_fired) {
+        opal_progress();
     }
-    
-    ORTE_PROGRESSED_WAIT(timer_fired, 0, 1);
 
     if (ORTE_SUCCESS != error_exit) {
         OBJ_DESTRUCT(&answer);
         return error_exit;
     }
-    
+
     cnt = 1;
     if (ORTE_SUCCESS != (ret = opal_dss.unpack(&answer, &cnt_jobs, &cnt, OPAL_INT32))) {
         ORTE_ERROR_LOG(ret);
         OBJ_DESTRUCT(&answer);
         return ret;
     }
-    
+
     /* allocate the required memory */
     if (0 < cnt_jobs) {
         job_info = (orte_job_t**)malloc(cnt_jobs * sizeof(orte_job_t*));
@@ -318,7 +344,7 @@ int orte_util_comm_query_job_info(const orte_process_name_t *hnp, orte_jobid_t j
         *num_jobs = cnt_jobs;
     }
     OBJ_DESTRUCT(&answer);
-    
+
     return ORTE_SUCCESS;
 }
 
@@ -330,11 +356,12 @@ int orte_util_comm_query_node_info(const orte_process_name_t *hnp, char *node,
     opal_buffer_t *cmd;
     orte_daemon_cmd_flag_t command = ORTE_DAEMON_REPORT_NODE_INFO_CMD;
     orte_node_t **node_info;
-    
+    struct timeval tv;
+
     /* set default response */
     *num_nodes = 0;
     *node_info_array = NULL;
-    
+
     /* query the HNP for node info */
     cmd = OBJ_NEW(opal_buffer_t);
     if (ORTE_SUCCESS != (ret = opal_dss.pack(cmd, &command, 1, ORTE_DAEMON_CMD))) {
@@ -347,25 +374,30 @@ int orte_util_comm_query_node_info(const orte_process_name_t *hnp, char *node,
         OBJ_RELEASE(cmd);
         return ret;
     }
-    /* define a max time to wait for send to complete */
+
+   /* define a max time to wait for send to complete */
     timer_fired = false;
     error_exit = ORTE_SUCCESS;
-    ORTE_DETECT_TIMEOUT(&quicktime, 100, 1000, 100000, quicktime_cb);
-    
+    quicktime = opal_event_alloc();
+    tv.tv_sec = 0;
+    tv.tv_usec = 100000;
+    opal_event_evtimer_set(orte_event_base, quicktime, quicktime_cb, NULL);
+    opal_event_set_priority(quicktime, ORTE_ERROR_PRI);
+    opal_event_evtimer_add(quicktime, &tv);
+
     /* do the send */
-    if (0 > (ret = orte_rml.send_buffer_nb((orte_process_name_t*)hnp, cmd, ORTE_RML_TAG_DAEMON, 0,
-                                           send_cbfunc, NULL))) {
+    if (0 > (ret = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                           (orte_process_name_t*)hnp, cmd,
+                                           ORTE_RML_TAG_DAEMON, send_cbfunc, NULL))) {
         ORTE_ERROR_LOG(ret);
         OBJ_RELEASE(cmd);
         return ret;
     }
-    
-    /* wait for send to complete */
-    ORTE_PROGRESSED_WAIT(timer_fired, 0, 1);
-    
-    /* release the buffer */
-    OBJ_RELEASE(cmd);
-    
+
+    while (!timer_fired) {
+        opal_progress();
+    }
+
     /* did it succeed? */
     if (ORTE_SUCCESS != error_exit) {
         return error_exit;
@@ -374,40 +406,39 @@ int orte_util_comm_query_node_info(const orte_process_name_t *hnp, char *node,
     /* define a max time to wait for an answer */
     timer_fired = false;
     error_exit = ORTE_SUCCESS;
-    ORTE_DETECT_TIMEOUT(&quicktime, 10, 1000, 10000, quicktime_cb);
-    
+
     /* get the answer */
     OBJ_CONSTRUCT(&answer, opal_buffer_t);
-    if (ORTE_SUCCESS != (ret = orte_rml.recv_buffer_nb(ORTE_NAME_WILDCARD,
-                                                      ORTE_RML_TAG_TOOL,
-                                                      ORTE_RML_NON_PERSISTENT,
-                                                      recv_info,
-                                                      NULL))) {
-        /* cancel the timer */
-        if (NULL != quicktime) {
-            opal_evtimer_del(quicktime);
-            free(quicktime);
-            quicktime = NULL;
-        }
-        ORTE_ERROR_LOG(ret);
-        OBJ_DESTRUCT(&answer);
-        return ret;
+    orte_rml.recv_buffer_nb(ORTE_NAME_WILDCARD,
+                            ORTE_RML_TAG_TOOL,
+                            ORTE_RML_NON_PERSISTENT,
+                            recv_info,
+                            NULL);
+
+    /* set a timer for getting the answer */
+    quicktime = opal_event_alloc();
+    tv.tv_sec = 0;
+    tv.tv_usec = 100000;
+    opal_event_evtimer_set(orte_event_base, quicktime, quicktime_cb, NULL);
+    opal_event_set_priority(quicktime, ORTE_ERROR_PRI);
+    opal_event_evtimer_add(quicktime, &tv);
+
+    while (!timer_fired) {
+        opal_progress();
     }
-    
-    ORTE_PROGRESSED_WAIT(timer_fired, 0, 1);
-    
+
     if (ORTE_SUCCESS != error_exit) {
         OBJ_DESTRUCT(&answer);
         return error_exit;
     }
-    
+
     cnt = 1;
     if (ORTE_SUCCESS != (ret = opal_dss.unpack(&answer, &cnt_nodes, &cnt, OPAL_INT32))) {
         ORTE_ERROR_LOG(ret);
         OBJ_DESTRUCT(&answer);
         return ret;
     }
-    
+
     /* allocate the required memory */
     if (0 < cnt_nodes) {
         node_info = (orte_node_t**)malloc(cnt_nodes * sizeof(orte_node_t*));
@@ -425,7 +456,7 @@ int orte_util_comm_query_node_info(const orte_process_name_t *hnp, char *node,
         *num_nodes = cnt_nodes;
     }
     OBJ_DESTRUCT(&answer);
-    
+
     return ORTE_SUCCESS;
 }
 
@@ -437,11 +468,13 @@ int orte_util_comm_query_proc_info(const orte_process_name_t *hnp, orte_jobid_t
     opal_buffer_t *cmd;
     orte_daemon_cmd_flag_t command = ORTE_DAEMON_REPORT_PROC_INFO_CMD;
     orte_proc_t **proc_info;
+    struct timeval tv;
+    char *nodename;
 
     /* set default response */
     *num_procs = 0;
     *proc_info_array = NULL;
-    
+
     /* query the HNP for info on the procs in this job */
     cmd = OBJ_NEW(opal_buffer_t);
     if (ORTE_SUCCESS != (ret = opal_dss.pack(cmd, &command, 1, ORTE_DAEMON_CMD))) {
@@ -459,25 +492,30 @@ int orte_util_comm_query_proc_info(const orte_process_name_t *hnp, orte_jobid_t
         OBJ_RELEASE(cmd);
         return ret;
     }
+
     /* define a max time to wait for send to complete */
     timer_fired = false;
     error_exit = ORTE_SUCCESS;
-    ORTE_DETECT_TIMEOUT(&quicktime, 100, 1000, 100000, quicktime_cb);
-    
+    quicktime = opal_event_alloc();
+    tv.tv_sec = 0;
+    tv.tv_usec = 100000;
+    opal_event_evtimer_set(orte_event_base, quicktime, quicktime_cb, NULL);
+    opal_event_set_priority(quicktime, ORTE_ERROR_PRI);
+    opal_event_evtimer_add(quicktime, &tv);
+
     /* do the send */
-    if (0 > (ret = orte_rml.send_buffer_nb((orte_process_name_t*)hnp, cmd, ORTE_RML_TAG_DAEMON, 0,
+    if (0 > (ret = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                           (orte_process_name_t*)hnp, cmd, ORTE_RML_TAG_DAEMON,
                                            send_cbfunc, NULL))) {
         ORTE_ERROR_LOG(ret);
         OBJ_RELEASE(cmd);
         return ret;
     }
-    
-    /* wait for send to complete */
-    ORTE_PROGRESSED_WAIT(timer_fired, 0, 1);
-    
-    /* release the buffer */
-    OBJ_RELEASE(cmd);
-    
+
+    while (!timer_fired) {
+        opal_progress();
+    }
+
     /* did it succeed? */
     if (ORTE_SUCCESS != error_exit) {
         return error_exit;
@@ -486,40 +524,39 @@ int orte_util_comm_query_proc_info(const orte_process_name_t *hnp, orte_jobid_t
     /* define a max time to wait for an answer */
     timer_fired = false;
     error_exit = ORTE_SUCCESS;
-    ORTE_DETECT_TIMEOUT(&quicktime, 10, 1000, 10000, quicktime_cb);
-    
+
     /* get the answer */
     OBJ_CONSTRUCT(&answer, opal_buffer_t);
-    if (ORTE_SUCCESS != (ret = orte_rml.recv_buffer_nb(ORTE_NAME_WILDCARD,
-                                                      ORTE_RML_TAG_TOOL,
-                                                      ORTE_RML_NON_PERSISTENT,
-                                                      recv_info,
-                                                      NULL))) {
-        /* cancel the timer */
-        if (NULL != quicktime) {
-            opal_evtimer_del(quicktime);
-            free(quicktime);
-            quicktime = NULL;
-        }
-        ORTE_ERROR_LOG(ret);
-        OBJ_DESTRUCT(&answer);
-        return ret;
+    orte_rml.recv_buffer_nb(ORTE_NAME_WILDCARD,
+                            ORTE_RML_TAG_TOOL,
+                            ORTE_RML_NON_PERSISTENT,
+                            recv_info,
+                            NULL);
+
+    /* set a timer for getting the answer */
+    quicktime = opal_event_alloc();
+    tv.tv_sec = 0;
+    tv.tv_usec = 100000;
+    opal_event_evtimer_set(orte_event_base, quicktime, quicktime_cb, NULL);
+    opal_event_set_priority(quicktime, ORTE_ERROR_PRI);
+    opal_event_evtimer_add(quicktime, &tv);
+
+    while (!timer_fired) {
+        opal_progress();
     }
-    
-    ORTE_PROGRESSED_WAIT(timer_fired, 0, 1);
-    
+
     if (ORTE_SUCCESS != error_exit) {
         OBJ_DESTRUCT(&answer);
         return error_exit;
     }
-    
+
     cnt = 1;
     if (ORTE_SUCCESS != (ret = opal_dss.unpack(&answer, &cnt_procs, &cnt, OPAL_INT32))) {
         ORTE_ERROR_LOG(ret);
         OBJ_DESTRUCT(&answer);
         return ret;
     }
-    
+
     /* allocate the required memory */
     if (0 < cnt_procs) {
         proc_info = (orte_proc_t**)malloc(cnt_procs * sizeof(orte_proc_t*));
@@ -532,6 +569,25 @@ int orte_util_comm_query_proc_info(const orte_process_name_t *hnp, orte_jobid_t
                 free(proc_info);
                 return ret;
             }
+            /* the vpid and nodename for this proc are no longer packed
+             * in the ORTE_PROC packing routines to save space for other
+             * uses, so we have to unpack them separately
+             */
+            cnt = 1;
+            if (ORTE_SUCCESS != (ret = opal_dss.unpack(&answer, &proc_info[n]->pid, &cnt, OPAL_PID))) {
+                ORTE_ERROR_LOG(ret);
+                OBJ_DESTRUCT(&answer);
+                free(proc_info);
+                return ret;
+            }
+            cnt = 1;
+            if (ORTE_SUCCESS != (ret = opal_dss.unpack(&answer, &nodename, &cnt, OPAL_STRING))) {
+                ORTE_ERROR_LOG(ret);
+                OBJ_DESTRUCT(&answer);
+                free(proc_info);
+                return ret;
+            }
+            orte_set_attribute(&proc_info[n]->attributes, ORTE_PROC_NODENAME, ORTE_ATTR_LOCAL, nodename, OPAL_STRING);
         }
         *proc_info_array = proc_info;
         *num_procs = (int)cnt_procs;
@@ -547,183 +603,208 @@ int orte_util_comm_query_proc_info(const orte_process_name_t *hnp, orte_jobid_t
  * to the plm_proxy tag! So we have to go another route to get this
  * request processed
  */
+static bool reply_waiting;
+
+static void comm_cbfunc(int status, orte_process_name_t* sender,
+                        opal_buffer_t* buffer, orte_rml_tag_t tag,
+                        void* cbdata)
+{
+    opal_buffer_t *buf = (opal_buffer_t*)cbdata;
+
+    if (NULL != buf) {
+        opal_dss.copy_payload(buf, buffer);
+    }
+    reply_waiting = false;
+}
+
 int orte_util_comm_spawn_job(const orte_process_name_t *hnp, orte_job_t *jdata)
 {
-    opal_buffer_t buf;
+    opal_buffer_t *buf;
     orte_daemon_cmd_flag_t command;
     orte_std_cntr_t count;
     int rc;
-    
+
     OPAL_OUTPUT_VERBOSE((5, orte_debug_output,
                          "%s util_comm_spawn_job: requesting HNP %s spawn new job",
                          ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
                          ORTE_NAME_PRINT(hnp)));
-    
+
     /* setup the buffer */
-    OBJ_CONSTRUCT(&buf, opal_buffer_t);
-    
+    buf = OBJ_NEW(opal_buffer_t);
+
     /* tell the HNP we are sending a launch request */
     command = ORTE_DAEMON_SPAWN_JOB_CMD;
-    if (ORTE_SUCCESS != (rc = opal_dss.pack(&buf, &command, 1, ORTE_DAEMON_CMD))) {
+    if (ORTE_SUCCESS != (rc = opal_dss.pack(buf, &command, 1, ORTE_DAEMON_CMD))) {
         ORTE_ERROR_LOG(rc);
+        OBJ_RELEASE(buf);
         goto CLEANUP;
     }
-    
+
     /* pack the jdata object */
-    if (ORTE_SUCCESS != (rc = opal_dss.pack(&buf, &jdata, 1, ORTE_JOB))) {
+    if (ORTE_SUCCESS != (rc = opal_dss.pack(buf, &jdata, 1, ORTE_JOB))) {
         ORTE_ERROR_LOG(rc);
+        OBJ_RELEASE(buf);
         goto CLEANUP;
-        
+
     }
-    
+
     OPAL_OUTPUT_VERBOSE((5, orte_debug_output,
                          "%s util_comm_spawn_job: sending spawn cmd to HNP %s",
                          ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
                          ORTE_NAME_PRINT(hnp)));
-    
+
     /* tell the target HNP to launch the job */
-    if (0 > (rc = orte_rml.send_buffer((orte_process_name_t*)hnp, &buf, ORTE_RML_TAG_DAEMON, 0))) {
+    if (0 > (rc = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                          (orte_process_name_t*)hnp, buf,
+                                          ORTE_RML_TAG_DAEMON,
+                                          orte_rml_send_callback, NULL))) {
         ORTE_ERROR_LOG(rc);
+        OBJ_RELEASE(buf);
         goto CLEANUP;
     }
-    OBJ_DESTRUCT(&buf);
-    
-    
+
+
     OPAL_OUTPUT_VERBOSE((5, orte_debug_output,
                          "%s util_comm_spawn_job: waiting for response",
                          ORTE_NAME_PRINT(ORTE_PROC_MY_NAME)));
-    
+
     /* wait for the target's response */
-    OBJ_CONSTRUCT(&buf, opal_buffer_t);
-    if (0 > (rc = orte_rml.recv_buffer(ORTE_NAME_WILDCARD, &buf, ORTE_RML_TAG_TOOL, 0))) {
-        ORTE_ERROR_LOG(rc);
-        goto CLEANUP;
-    }
-    
+    reply_waiting = true;
+    buf = OBJ_NEW(opal_buffer_t);
+    orte_rml.recv_buffer_nb(ORTE_NAME_WILDCARD, ORTE_RML_TAG_TOOL, 0,
+                            comm_cbfunc, buf);
+    ORTE_WAIT_FOR_COMPLETION(reply_waiting);
+
     /* get the new jobid back in case the caller wants it */
     count = 1;
-    if (ORTE_SUCCESS != (rc = opal_dss.unpack(&buf, &(jdata->jobid), &count, ORTE_JOBID))) {
+    if (ORTE_SUCCESS != (rc = opal_dss.unpack(buf, &(jdata->jobid), &count, ORTE_JOBID))) {
         ORTE_ERROR_LOG(rc);
+        OBJ_RELEASE(buf);
         goto CLEANUP;
     }
     if (ORTE_JOBID_INVALID == jdata->jobid) {
         /* something went wrong on far end - go no further */
         rc = ORTE_ERR_FAILED_TO_START;
+        OBJ_RELEASE(buf);
         goto CLEANUP;
     }
-    
+    OBJ_RELEASE(buf);
+
     /* good to go! */
-    
+
 CLEANUP:
-    OBJ_DESTRUCT(&buf);
-    
     return rc;
 }
 
 
 int orte_util_comm_terminate_job(const orte_process_name_t *hnp, orte_jobid_t job)
 {
-    opal_buffer_t buf;
+    opal_buffer_t *buf;
     orte_daemon_cmd_flag_t command;
     orte_std_cntr_t count;
     int rc, ret = ORTE_ERROR;
-    
+
     OPAL_OUTPUT_VERBOSE((5, orte_debug_output,
                          "%s util_comm_spawn_job: requesting HNP %s terminate job %s",
                          ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
                          ORTE_NAME_PRINT(hnp),
                          ORTE_JOBID_PRINT(job)));
-    
+
     /* setup the buffer */
-    OBJ_CONSTRUCT(&buf, opal_buffer_t);
-    
+    buf = OBJ_NEW(opal_buffer_t);
+
     /* tell the HNP we are sending a terminate request */
     command = ORTE_DAEMON_TERMINATE_JOB_CMD;
-    if (ORTE_SUCCESS != (rc = opal_dss.pack(&buf, &command, 1, ORTE_DAEMON_CMD))) {
+    if (ORTE_SUCCESS != (rc = opal_dss.pack(buf, &command, 1, ORTE_DAEMON_CMD))) {
         ORTE_ERROR_LOG(rc);
         ret = rc;
+        OBJ_RELEASE(buf);
         goto CLEANUP;
     }
-    
+
     /* pack the jobid */
-    if (ORTE_SUCCESS != (rc = opal_dss.pack(&buf, &job, 1, ORTE_JOBID))) {
+    if (ORTE_SUCCESS != (rc = opal_dss.pack(buf, &job, 1, ORTE_JOBID))) {
         ORTE_ERROR_LOG(rc);
         ret = rc;
+        OBJ_RELEASE(buf);
         goto CLEANUP;
     }
-    
+
     OPAL_OUTPUT_VERBOSE((5, orte_debug_output,
                          "%s util_comm_spawn_job: sending terminate cmd to HNP %s",
                          ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
                          ORTE_NAME_PRINT(hnp)));
-    
+
     /* tell the target HNP to terminate the job */
-    if (0 > (rc = orte_rml.send_buffer((orte_process_name_t*)hnp, &buf, ORTE_RML_TAG_DAEMON, 0))) {
+    if (0 > (rc = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                          (orte_process_name_t*)hnp, buf,
+                                          ORTE_RML_TAG_DAEMON,
+                                          orte_rml_send_callback, NULL))) {
         ORTE_ERROR_LOG(rc);
         ret = rc;
+        OBJ_RELEASE(buf);
         goto CLEANUP;
     }
-    OBJ_DESTRUCT(&buf);
-    
-    
+
+
     OPAL_OUTPUT_VERBOSE((5, orte_debug_output,
                          "%s util_comm_terminate_job: waiting for response",
                          ORTE_NAME_PRINT(ORTE_PROC_MY_NAME)));
-    
+
     /* wait for the target's response */
-    OBJ_CONSTRUCT(&buf, opal_buffer_t);
-    if (0 > (rc = orte_rml.recv_buffer(ORTE_NAME_WILDCARD, &buf, ORTE_RML_TAG_TOOL, 0))) {
-        ORTE_ERROR_LOG(rc);
-        ret = rc;
-        goto CLEANUP;
-    }
-    
+    reply_waiting = true;
+    buf = OBJ_NEW(opal_buffer_t);
+    orte_rml.recv_buffer_nb(ORTE_NAME_WILDCARD,
+                            ORTE_RML_TAG_TOOL, 0,
+                            comm_cbfunc, buf);
+    ORTE_WAIT_FOR_COMPLETION(reply_waiting);
+
     /* get the status code */
     count = 1;
-    if (ORTE_SUCCESS != (rc = opal_dss.unpack(&buf, &ret, &count, OPAL_INT))) {
+    if (ORTE_SUCCESS != (rc = opal_dss.unpack(buf, &ret, &count, OPAL_INT))) {
         ORTE_ERROR_LOG(rc);
         ret = rc;
-        goto CLEANUP;
     }
-    
+    OBJ_RELEASE(buf);
+
 CLEANUP:
-    OBJ_DESTRUCT(&buf);
-    
     return ret;
 }
 
 int orte_util_comm_halt_vm(const orte_process_name_t *hnp)
 {
-    opal_buffer_t buf;
+    opal_buffer_t *buf;
     orte_daemon_cmd_flag_t command;
     int rc;
-    
+
     OPAL_OUTPUT_VERBOSE((5, orte_debug_output,
                          "%s util_comm_halt_vm: ordering HNP %s terminate",
                          ORTE_NAME_PRINT(ORTE_PROC_MY_NAME),
                          ORTE_NAME_PRINT(hnp)));
-    
+
     /* setup the buffer */
-    OBJ_CONSTRUCT(&buf, opal_buffer_t);
-    
+    buf = OBJ_NEW(opal_buffer_t);
+
     /* tell the HNP to die */
     command = ORTE_DAEMON_HALT_VM_CMD;
-    if (ORTE_SUCCESS != (rc = opal_dss.pack(&buf, &command, 1, ORTE_DAEMON_CMD))) {
+    if (ORTE_SUCCESS != (rc = opal_dss.pack(buf, &command, 1, ORTE_DAEMON_CMD))) {
         ORTE_ERROR_LOG(rc);
+        OBJ_RELEASE(buf);
         goto CLEANUP;
     }
-    
+
     /* send the order */
-    if (0 > (rc = orte_rml.send_buffer((orte_process_name_t*)hnp, &buf, ORTE_RML_TAG_DAEMON, 0))) {
+    if (0 > (rc = orte_rml.send_buffer_nb(orte_mgmt_conduit,
+                                          (orte_process_name_t*)hnp, buf,
+                                          ORTE_RML_TAG_DAEMON,
+                                          orte_rml_send_callback, NULL))) {
         ORTE_ERROR_LOG(rc);
+        OBJ_RELEASE(buf);
         goto CLEANUP;
     }
-    OBJ_DESTRUCT(&buf);
-    
+    OBJ_RELEASE(buf);
+
     /* don't bother waiting around */
 CLEANUP:
-    OBJ_DESTRUCT(&buf);
-    
     return rc;
 }
-
